/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "compiler.h"
#include "nvme_emu.h"
#include "nvme_controller.h"

#define NVME_DEBUG_MAX_DUMP_FUNC_LEN 256
#define NVME_DEBUG_MAX_CMD_DUMP_LEN 128

static char *nvme_adm_cmd_feature_str(int feature)
{
    switch (feature) {
        case NVME_ARBITRATION:
            return "arbitration";
        case NVME_POWER_MANAGEMENT:
            return "pwr mgmt";
        case NVME_LBA_RANGE_TYPE:
            return "lba_range_type";
        case NVME_TEMPERATURE_THRESHOLD:
            return "temp_thresh";
        case NVME_ERROR_RECOVERY:
            return "error_recovery";
        case NVME_VOLATILE_WRITE_CACHE:
            return "write_cache";
        case NVME_NUMBER_OF_QUEUES:
            return "num_queues";
        case NVME_INTERRUPT_COALESCING:
            return "intr_coalescing";
        case NVME_INTERRUPT_VECTOR_CONF:
            return "intr_vector_conf";
        case NVME_WRITE_ATOMICITY:
            return "write_atomicity";
        case NVME_ASYNCHRONOUS_EVENT_CONF:
            return "async_ev_conf";
        case NVME_SOFTWARE_PROGRESS_MARKER:
            return "software_progress_marker";
        default:
            return "unknown_feature";
    }
}

static char *nvme_adm_cmd_feature_val(uint8_t ftr_val)
{
    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        return "current";
    case NVME_FEATURE_DEFAULT:
        return "default";
    case NVME_FEATURE_SAVED:
        return "saved";
    case NVME_FEATURE_CAPS:
        return "caps";
    }
    return "bad_feature_select";
}

static char *nvme_admin_cmd_str(int opcode)
{
    switch (opcode) {
        case NVME_ADMIN_DELETE_SQ:
            return "DEL_SQ";
        case NVME_ADMIN_CREATE_SQ:
            return "CREAT_SQ";
        case NVME_ADMIN_GET_LOG_PAGE:
            return "GET_LOG_PAGE";
        case NVME_ADMIN_DELETE_CQ:
            return "DEL_CQ";
        case NVME_ADMIN_CREATE_CQ:
            return "CREAT_CQ";
        case NVME_ADMIN_IDENTIFY:
            return "IDENTIFY";
        case NVME_ADMIN_ABORT:
            return "ABORT";
        case NVME_ADMIN_SET_FEATURES:
            return "SET_FTRS";
        case NVME_ADMIN_GET_FEATURES:
            return "GET_FTRS";
        case NVME_ADMIN_ASYNC_EV_REQ:
            return "ASYNC_EV_REQ";
        case NVME_ADMIN_FW_COMMIT:
            return "FW_COMMIT";
        case NVME_ADMIN_FW_DOWNLOAD:
            return "FW_DOWNLOAD";
        case NVME_ADMIN_FORMAT_NVM:
            return "FMT_NVM";
        case NVME_ADMIN_SECURITY_SEND:
            return "SCRT_SEND";
        case NVME_ADMIN_SECURITY_RECV:
            return "SCRT_RECV";
        case NVME_ADMIN_FW_RECOVER:
            return "FW_RECOVER";
        case NVME_ADMIN_VS_JSON_RPC_2_0_REQ:
            return "VENDOR_SPECIFIC_JSON_RPC_2_0_REQUEST";
        case NVME_ADMIN_VS_JSON_RPC_2_0_RSP:
            return "VENDOR_SPECIFIC_JSON_RPC_2_0_RESPONSE";
        case NVME_ADMIN_VS_IOVA_MGMT:
            return "NVME_ADMIN_VS_IOVA_MGMT";
        default:
            return "UNKNOWN";
    }
};

/* dword0 dump */
void nvme_cmd_dump_dword0(nvme_cmd_t *cmd, char *dump)
{
    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "FUSE:%d SGL:%d ", cmd->common.fuse , cmd->common.psdt);
}

void nvme_adm_cmd_identify_dump(const nvme_cmd_t *cmd, char *dump)
{
   const nvme_cmd_identify_t *icmd = (nvme_cmd_identify_t *)cmd;
   char cns_dump[NVME_DEBUG_MAX_CMD_DUMP_LEN] = {0};

   switch (le32_to_cpu(icmd->cns)) {
       case NVME_IDENTIFY_CNS_NAMESPACE:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "id_ns %d", le32_to_cpu(icmd->common.nsid));
           break;
       case NVME_IDENTIFY_CNS_CTRL:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "id_ctrl");
           break;
       case NVME_IDENTIFY_CNS_ACTIVE_NS_LIST:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "nslist");
           break;
       case NVME_IDENTIFY_CNS_NS_ID_DESCRIPTOR_LIST:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "ns_id_descriptor_list");
           break;
       case NVME_IDENTIFY_CNS_ALLOCATED_NS_LIST:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "allocated_nslist");
           break;
       default:
           snprintf(cns_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                    "unsupported");
   }
   snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
            "%s prp1:0x%llx prp2:0x%llx",
            cns_dump, (unsigned long long)icmd->common.dptr.prp1,
            (unsigned long long)icmd->common.dptr.prp2);
}

void nvme_adm_abort_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_abort_t *c = (nvme_cmd_abort_t *)cmd;
    uint16_t abort_cid = le16_to_cpu(c->cid);
    uint16_t abort_sqid = le16_to_cpu(c->sqid);

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "abort cmd with id:%d in queue:%d",
             abort_cid, abort_sqid);
}

void nvme_adm_cmd_create_cq_dump(const nvme_cmd_t *cmd, char *dump)
{
   const nvme_cmd_create_cq_t *c = (nvme_cmd_create_cq_t *)cmd;

   uint16_t cqid = le16_to_cpu(c->qid);
   uint16_t vector = le16_to_cpu(c->iv);
   uint16_t qsize = le16_to_cpu(c->qsize);
   uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);

   snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
            "base:%p id:%d size:%d int_v:0x%x pc:%d ien:%d",
            (void *)(uintptr_t)prp1, cqid, qsize, vector, c->pc, c->ien);
}

void nvme_adm_cmd_create_sq_dump(const nvme_cmd_t *cmd, char *dump)
{
   const nvme_cmd_create_sq_t *c = (nvme_cmd_create_sq_t *)cmd;

   uint16_t cqid = le16_to_cpu(c->cqid);
   uint16_t sqid = le16_to_cpu(c->qid);
   uint16_t qsize = le16_to_cpu(c->qsize);
   uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);

   snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
           "base:%p id:%d cqid:%d size:%d pc:%d qprio:%d",
           (void *)(uintptr_t)prp1, sqid, cqid, qsize, c->pc, c->qprio);
}

void nvme_adm_cmd_delete_q_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_delete_q_t *c = (nvme_cmd_delete_q_t *)cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
            "qid:%d", le16_to_cpu(c->qid));
}

void nvme_adm_cmd_set_features_dump(const nvme_cmd_t *cmd, char *dump)
{
    uint32_t dw10 = le32_to_cpu(cmd->cdw10);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);
    char dw10_dump[NVME_DEBUG_MAX_CMD_DUMP_LEN] = {0};
    
    switch (dw10) {
    case NVME_NUMBER_OF_QUEUES:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "sq:%d cq:%d", (dw11 & 0xFFFF), (dw11>>16) & 0xFFFF);
        break;
    case NVME_ARBITRATION:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "burst: %d, lpw: %d, mpw: %d, hpw: %d",
                 (dw11 & 0x7), ((dw11 >> 8) & 0xFF), ((dw11 >> 16) & 0xFF),
                 ((dw11 >> 24) & 0xFF));
        break;
    case NVME_ERROR_RECOVERY:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 " dulbe: %d, tler: %d", (dw11>>16) & 0x1, (dw11 & 0xFFFF));
        break;
    case NVME_INTERRUPT_COALESCING:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "aggregation threshold %u, aggregation time %u",
                  (dw11 & 0xFF), ((dw11 >> 8) & 0xFF));
        break;
    case NVME_ASYNCHRONOUS_EVENT_CONF:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "dw11: 0x%x", dw11);
        break;
    case NVME_POWER_MANAGEMENT:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 " wh: 0x%x ps: %d", (dw11 >> 4) & 0x7, dw11 & 0xf);
        break;
    default:
        snprintf(dw10_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "unsupported");
    }
    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "%s: 0x%x %s", nvme_adm_cmd_feature_str(dw10), dw10,
             dw10_dump);
}

void nvme_acmd_cmd_get_feature_dump(const nvme_cmd_t *cmd, char *dump)
{
    uint32_t dw10 = le32_to_cpu(cmd->cdw10);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);
    char ftr_id_dump[NVME_DEBUG_MAX_CMD_DUMP_LEN] = {0};

    const nvme_cmd_get_feature_t *c = (nvme_cmd_get_feature_t *)cmd;

    switch (c->fid) {
    case NVME_ARBITRATION:
        break;
    case NVME_POWER_MANAGEMENT:
        break;
    case NVME_TEMPERATURE_THRESHOLD:
        snprintf(ftr_id_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                "%s threshold", (dw11 & (3 << 20)) ? "under" : "over");
        break;
    case NVME_ERROR_RECOVERY:
        break;
    case NVME_NUMBER_OF_QUEUES:
        break;
    case NVME_INTERRUPT_COALESCING:
        break;
    case NVME_INTERRUPT_VECTOR_CONF:
        break;
    case NVME_WRITE_ATOMICITY:
        break;
    case NVME_ASYNCHRONOUS_EVENT_CONF:
        break;
    default:
        snprintf(ftr_id_dump, NVME_DEBUG_MAX_CMD_DUMP_LEN,
                 "unsupported");
    }

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "%s: val %s %s", nvme_adm_cmd_feature_str(dw10),
             nvme_adm_cmd_feature_val(c->sel),
             ftr_id_dump);
}

static char *log_page_name(int id)
{
    switch(id) {
        case NVME_LOG_ERROR_INFO:
            return "error_log";
        case NVME_LOG_SMART_INFO:
            return "smart_log";
        case NVME_LOG_FW_SLOT_INFO:
            return "fw_log";
        case NVME_LOG_CHANGED_NS_LIST_INFO:
            return "changed_nslist_log";
        case NVME_LOG_COMMANDS_EFFECTS_LOG:
            return "effects_log";
        case NVME_LOG_BF_PKG_INFO:
            return "bf_pkg_log";
        default:
            return "unknown";
    }
}

void nvme_adm_cmd_get_log_page_dump(const nvme_cmd_t *cmd, char *dump)
{
    uint32_t dw10 = le32_to_cpu(cmd->cdw10);
    const nvme_cmd_get_page_log_t *c = (nvme_cmd_get_page_log_t *)cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "page id %u (%s), num dwords %u, 0x%x  prp1:0x%llx prp2:0x%llx rae:%d",
             c->lid, log_page_name(c->lid),
             nvme_acmd_get_log_page_cmd_num_dwords(c), dw10,
             (unsigned long long)cmd->common.dptr.prp1, (unsigned long long)cmd->common.dptr.prp2,
             c->rae);
}

static void nvme_adm_fw_download_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_fw_download_t *c = (const nvme_cmd_fw_download_t*) cmd;

    const uint32_t numd = le32_to_cpu(c->numd) + 1;
    const uint32_t offset = le32_to_cpu(c->ofst);

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
            "num dwords: 0x%x, offset: 0x%x, prp1:0x%lx prp2:0x%lx", numd,
            offset, (uint64_t) cmd->common.dptr.prp1, (uint64_t) cmd->common.dptr.prp2);
}

static void nvme_adm_fw_commit_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_fw_commit_t *c = (const nvme_cmd_fw_commit_t*) cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
            "Firmware slot: %u, commit action: %u, boot partition id %u",
            c->firmware_slot, c->commit_action, c->bpid);
}

static void nvme_adm_fw_recover_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_fw_recover_t *c = (const nvme_cmd_fw_recover_t*) cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN, "Package Identifier: %d", c->pid);
}

void nvme_adm_cmd_vendor_specific_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_vendor_specific_cmd_t *c = (nvme_vendor_specific_cmd_t *)cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "num dwords %u, ndm %u  prp1:0x%llx prp2:0x%llx",
             le32_to_cpu(c->ndt), le32_to_cpu(c->ndm),
             (unsigned long long)cmd->common.dptr.prp1, (unsigned long long)cmd->common.dptr.prp2);
}

void nvme_adm_cmd_iova_mgmt_dump(const nvme_cmd_t *cmd, char *dump)
{
    const nvme_cmd_vs_iova_mgmt_t *c = (nvme_cmd_vs_iova_mgmt_t *)cmd;

    snprintf(dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
             "opm: %u, fid: %d, siova: 0x%lx, tiova: 0x%lx, size: 0x%x, szu: %u",
             c->opm, (int32_t)c->fid,
             c->siova, c->tiova, c->size, c->szu);
}

void nvme_admin_cmd_dump(const nvme_cmd_t *cmd)
{
    char cmd_dump[NVME_DEBUG_MAX_DUMP_FUNC_LEN] = {0};

    switch (cmd->common.opc) {
        case NVME_ADMIN_IDENTIFY:
            nvme_adm_cmd_identify_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_ABORT:
            nvme_adm_abort_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_CREATE_CQ:
            nvme_adm_cmd_create_cq_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_DELETE_CQ:
            nvme_adm_cmd_delete_q_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_CREATE_SQ:
            nvme_adm_cmd_create_sq_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_DELETE_SQ:
            nvme_adm_cmd_delete_q_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_SET_FEATURES:
            nvme_adm_cmd_set_features_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_GET_FEATURES:
            nvme_acmd_cmd_get_feature_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_GET_LOG_PAGE:
            nvme_adm_cmd_get_log_page_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_ASYNC_EV_REQ:
            break;
        case NVME_ADMIN_FW_DOWNLOAD:
            nvme_adm_fw_download_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_FW_COMMIT:
            nvme_adm_fw_commit_dump(cmd, cmd_dump);
            break;
        case NVME_ADMIN_FW_RECOVER:
            nvme_adm_fw_recover_dump(cmd, cmd_dump);
            break;
#if HAVE_SNAP
        case NVME_ADMIN_VS_JSON_RPC_2_0_REQ:
            /* fallthrough */
        case NVME_ADMIN_VS_JSON_RPC_2_0_RSP:
            nvme_adm_cmd_vendor_specific_dump(cmd, cmd_dump);
            break;
#endif
        case NVME_ADMIN_VS_IOVA_MGMT:
            nvme_adm_cmd_iova_mgmt_dump(cmd, cmd_dump);
            break;
        default:
            snprintf(cmd_dump, NVME_DEBUG_MAX_DUMP_FUNC_LEN,
                     "UNSUPPORTED");
    }
    nvmx_info("ACMD: 0x%x %s tag: %d %s", cmd->common.opc,
              nvme_admin_cmd_str(cmd->common.opc), cmd->common.cid,
              cmd_dump);
}

static char *nvme_cmd_str(int opcode)
{
    switch (opcode) {
        case NVME_IO_FLUSH:
            return "FLUSH";
        case NVME_IO_WRITE:
            return "WRITE";
        case NVME_IO_READ:
            return "READ";
        case NVME_IO_WRITE_UNCOR:
            return "WRITE_UNCOR";
        case NVME_IO_COMPARE:
            return "COMPARE";
        case NVME_IO_WRITE_ZEROS:
            return "WRITE_ZEROES";
        case NVME_IO_DSM:
            return "DSM";
        default:
            return "UNKNOWN";
    }
}

static void nvme_io_cmd_rw_dump(const nvme_cmd_t *cmd)
{
    const nvme_cmd_rw_t *rw = (nvme_cmd_rw_t *)cmd;

    if (!rw)
        return;

    switch (cmd->common.psdt) {
    case NVME_PSDT_PRP:
        nvmx_debug("cid:%d prp1:%p prp2:%p start_blk:%llu nblks:%lu", rw->common.cid,
                  (void *)(uintptr_t)le64_to_cpu(rw->common.dptr.prp1),
                  (void *)(uintptr_t)le64_to_cpu(rw->common.dptr.prp2),
                  (unsigned long long)le64_to_cpu(rw->slba),
                  (unsigned long)(le32_to_cpu(rw->nlb) + 1));
        break;

    case NVME_PSDT_SGL_MPTR_CONTIG:
        nvmx_debug("cid:%d sgl_type:%u sgl_subtype:%u start_blk:%llu nblks:%lu", rw->common.cid,
                  (unsigned int)le32_to_cpu(rw->common.dptr.sgl.type),
                  (unsigned int)le32_to_cpu(rw->common.dptr.sgl.subtype),
                  (unsigned long long)le64_to_cpu(rw->slba),
                  (unsigned long)(le32_to_cpu(rw->nlb) + 1));
        break;

    default:
        break;
    }
}

static void nvme_io_cmd_dsm_dump(const nvme_cmd_t *cmd)
{
    const nvme_cmd_dsm_t *dsm = (nvme_cmd_dsm_t *)cmd;

    nvmx_info("cid:%d prp1:%p prp2:%p nr %u attributes %u", dsm->common.cid,
              (void *)(uintptr_t)le64_to_cpu(dsm->common.dptr.prp1),
              (void *)(uintptr_t)le64_to_cpu(dsm->common.dptr.prp2),
              dsm->nr, dsm->attributes);
}

void nvme_io_cmd_dump(const nvme_cmd_t *cmd)
{
    nvmx_debug("IOCMD: 0x%x %s ns:%d tag:%d", cmd->common.opc,
               nvme_cmd_str(cmd->common.opc), le32_to_cpu(cmd->common.nsid), cmd->common.cid);

    switch (cmd->common.opc) {
        case NVME_IO_READ:
        case NVME_IO_WRITE:
            nvme_io_cmd_rw_dump(cmd);
            break;
        case NVME_IO_FLUSH:
            return;
        case NVME_IO_DSM:
            nvme_io_cmd_dsm_dump(cmd);
            break;
        default:
            nvmx_warn("IOCMD: 0x%x %s ns:%d tag:%d UNSUPPORTED",
                      cmd->common.opc, nvme_cmd_str(cmd->common.opc),
                      le32_to_cpu(cmd->common.nsid), cmd->common.cid);
    }
}

