/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <sys/syscall.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include <execinfo.h>
#include <stdarg.h>

#include "config.h"
#include "nvme_emu_log.h"

#define NVMX_SLEEP_ON_FATAL_VAR "NVMX_SLEEP_ON_FATAL"
#define NVMX_BACKTRACE_MAX    64

#define NVMX_DEBUG_PRINT_STRING(_out, _str, ...) \
    do { \
        if (!(_out)) { \
            nvmx_info((_str), ## __VA_ARGS__); \
        } else { \
            fprintf((_out), (_str), ## __VA_ARGS__); \
        } \
    } while(0);

static FILE *out;
static bool sig_dfl;
static bool initialized;
static void *debug_signal_restorer;

static void nvmx_debug_freeze()
{
    nvmx_debug("thread %lu is freezing", pthread_self());
    pause();
}

static void nvmx_debug_stop_handler(int signo)
{
    nvmx_debug_freeze();
}

static void nvmx_debug_stop_other_threads()
{
    static const char *task_dir = "/proc/self/task";
    struct dirent *entry;
    DIR *dir;
    int ret;
    int tid;
    struct sigaction act;

    dir = opendir(task_dir);
    if (dir == NULL) {
        nvmx_error("Unable to open %s: %m\n", task_dir);
        return;
    }

    memset(&act, 0, sizeof(act));
    act.sa_handler = nvmx_debug_stop_handler;
    sigaction(SIGUSR1, &act, 0);

    for (;;) {
        errno = 0;
        entry = readdir(dir);
        if (entry == NULL) {
            if (errno != 0) {
                nvmx_error("Unable to read from %s: %m", task_dir);
            }
            break;
        }

        if (!strncmp(entry->d_name, ".", 1)) {
            continue;
        }

        tid = atoi(entry->d_name);
        if ((tid == 0) || (tid == syscall(SYS_gettid))) {
            continue;
        }

        ret = syscall(SYS_tgkill, getpid(), tid, SIGUSR1);
        if (ret < 0) {
             break;
        }
    }

    closedir(dir);
}

void nvmx_error_freeze()
{
    if (getenv(NVMX_SLEEP_ON_FATAL_VAR)) {
        nvmx_warn("NVMX application entered freeze state");
        nvmx_debug_stop_other_threads();
        nvmx_debug_freeze();
    }
}

static const char *signal_cause_common(int si_code)
{
    switch (si_code) {
        case SI_USER      : return "kill(2) or raise(3)";
        case SI_KERNEL    : return "Sent by the kernel";
        case SI_QUEUE     : return "sigqueue(2)";
        case SI_TIMER     : return "POSIX timer expired";
        case SI_MESGQ     : return "POSIX message queue state changed";
        case SI_ASYNCIO   : return "AIO completed";
        case SI_SIGIO     : return "queued SIGIO";
        case SI_TKILL     : return "tkill(2) or tgkill(2)";
        default           : return "<unknown si_code>";
    }
}

static const char *signal_cause_ill(int si_code)
{
    switch (si_code) {
        case ILL_ILLOPC   : return "illegal opcode";
        case ILL_ILLOPN   : return "illegal operand";
        case ILL_ILLADR   : return "illegal addressing mode";
        case ILL_ILLTRP   : return "illegal trap";
        case ILL_PRVOPC   : return "privileged opcode";
        case ILL_PRVREG   : return "privileged register";
        case ILL_COPROC   : return "coprocessor error";
        case ILL_BADSTK   : return "internal stack error";
        default           : return signal_cause_common(si_code);
    }
}

static const char *signal_cause_segv(int si_code)
{
    switch (si_code) {
        case SEGV_MAPERR  : return "address not mapped to object";
        case SEGV_ACCERR  : return "invalid permissions for mapped object";
        default           : return signal_cause_common(si_code);
    }
}

static const char *signal_cause_bus(int si_code)
{
    switch (si_code) {
        case BUS_ADRALN   : return "invalid address alignment";
        case BUS_ADRERR   : return "nonexistent physical address";
        case BUS_OBJERR   : return "object-specific hardware error";
        default           : return signal_cause_common(si_code);
    }
}

static const char *signal_cause_trap(int si_code)
{
    switch (si_code) {
#ifdef TRAP_BRKPT
        case TRAP_BRKPT   : return "process breakpoint";
#endif
#ifdef TRAP_TRACE
        case TRAP_TRACE   : return "process trace trap";
#endif
        default           : return signal_cause_common(si_code);
    }
}

static int debug_backtrace_is_excluded(void *address, const char *symbol)
{
    return (address == debug_signal_restorer);
}

void debug_print_backtrace()
{
    char **symbols;
    void *addresses[NVMX_BACKTRACE_MAX];
    int count, i, n;

    if (!initialized)
        return;

    NVMX_DEBUG_PRINT_STRING(out, "==== backtrace ====");

    count = backtrace(addresses, NVMX_BACKTRACE_MAX);
    symbols = backtrace_symbols(addresses, count);

    n = 0;
    for (i = 0; i < count; ++i) {
        if (!debug_backtrace_is_excluded(addresses[i], symbols[i])) {
            NVMX_DEBUG_PRINT_STRING(out, "   %2d  %s", n, symbols[i]);
            ++n;
        }
    }

    free(symbols);

    NVMX_DEBUG_PRINT_STRING(out, "===================");
}

static int set_signal_handler(void (*handler)(int, siginfo_t*, void *))
{
    int ret;
    struct sigaction sigact;
    struct sigaction old_action;

    memset(&sigact, 0, sizeof(sigact));
    memset(&old_action, 0, sizeof(old_action));

    sigact.sa_sigaction = handler;
    sigact.sa_flags     = SA_SIGINFO;

    sigemptyset(&sigact.sa_mask);

    ret = sigaction(SIGILL, &sigact, NULL);
    if (ret < 0) {
        nvmx_error("failed to set signal handler for signal SIGILL : %m");
        goto sigill_error;
    }

    ret = sigaction(SIGTRAP, &sigact, NULL);
    if (ret < 0) {
        nvmx_error("failed to set signal handler for signal SIGTRAP : %m");
        goto sigtrap_error;
    }

    ret = sigaction(SIGBUS, &sigact, NULL);
    if (ret < 0) {
        nvmx_error("failed to set signal handler for signal SIGBUS : %m");
        goto sigbus_error;
    }

    ret = sigaction(SIGSEGV, &sigact, &old_action);
    if (ret < 0) {
        nvmx_error("failed to set signal handler for signal SIGSEGV : %m");
        goto sigsegv_error;
    }

    /*
     * The signal comes to glibc form the Kernel always from the same address,
     * remember this address in order to skip it while printing.
     */
    debug_signal_restorer = old_action.sa_restorer;

    return 0;

sigsegv_error:
    signal(SIGBUS, SIG_DFL);
sigbus_error:
    signal(SIGTRAP, SIG_DFL);
sigtrap_error:
    signal(SIGILL, SIG_DFL);
sigill_error:
    return ret;
}

static void debug_handle_error_signal(int signo, const char *cause,
                                          const char *fmt, ...)
{
    char buf[256];
    va_list ap;

    /* Note: you should avoid dynamic memory allocation in signal handler.
     * For more information please refer to redmine issue #3100194.
     */

    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);

    nvmx_error("Caught signal %d (%s: %s%s)", signo, strsignal(signo),
               cause, buf);

    debug_print_backtrace();
}

static void error_signal_handler(int signo, siginfo_t *info, void *context)
{
    switch (signo) {
    case SIGILL:
        debug_handle_error_signal(signo, signal_cause_ill(info->si_code), "");
        break;
    case SIGTRAP:
        debug_handle_error_signal(signo, signal_cause_trap(info->si_code), "");
        break;
    case SIGBUS:
        debug_handle_error_signal(signo, signal_cause_bus(info->si_code), "");
        break;
    case SIGSEGV:
        debug_handle_error_signal(signo, signal_cause_segv(info->si_code),
                                  " at address %p", info->si_addr);
        break;
    default:
        break;
    }

    if (sig_dfl) {
        signal(signo, SIG_DFL);
        raise(signo);
    }
}

void backtrace_stop()
{
    if (initialized) {
        out = NULL;
        sig_dfl = false;
        initialized = false;
        signal(SIGILL, SIG_DFL);
        signal(SIGBUS, SIG_DFL);
        signal(SIGTRAP, SIG_DFL);
        signal(SIGILL, SIG_DFL);
    }
}

int backtrace_init(FILE *stream, bool act)
{
    int ret;

    if (initialized)
        return -EALREADY;

    ret = set_signal_handler(error_signal_handler);
    if (ret < 0)
        return ret;

    out = stream;
    sig_dfl = act;
    initialized = true;

    return 0;
}
