/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdint.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <sys/queue.h>
#include "ref_ptr.h"
#include "nvme_emu_log.h"
#include "effdma_domain.h"

struct effdma_domain
{
    int pf_id;
    int vf_id;
    ref_ptr_t address_space;
    struct address_space *next_address_space;
    LIST_ENTRY(effdma_domain) entries;
};

LIST_HEAD(domains_head, effdma_domain);

static struct domains_head domains = LIST_HEAD_INITIALIZER();

void effdma_domain_clear()
{
    struct address_space *address_space;
    struct effdma_domain *item, *next_item;

    item = LIST_FIRST(&domains);
    while (item != NULL) {
        next_item = LIST_NEXT(item, entries);
        address_space = (struct address_space *)ref_ptr(&item->address_space);
        address_space_destroy(address_space);
        free(item);
        item = next_item;
    }
    LIST_INIT(&domains);
}

struct effdma_domain *effdma_domain_find(int pf_id, int vf_id)
{
    struct effdma_domain *item;

    LIST_FOREACH(item, &domains, entries) {
        if (item->pf_id == pf_id && item->vf_id == vf_id)
            return item;
    }

    return NULL;
}

struct effdma_domain *effdma_domain_find_or_create(int pf_id, int vf_id)
{
    struct effdma_domain *item;

    item = effdma_domain_find(pf_id, vf_id);
    if (!item) {
        item = calloc(1, sizeof(struct effdma_domain));
        if (item) {
            item->pf_id = pf_id;
            item->vf_id = vf_id;
            ref_ptr_init(&item->address_space, address_space_init());
            if (ref_ptr(&item->address_space))
                LIST_INSERT_HEAD(&domains, item, entries);
            else {
                free(item);
                item = NULL;
            }
        }
    }

    return item;
}

address_space_t *effdma_domain_transaction_begin(struct effdma_domain *domain)
{
    domain->next_address_space = address_space_clone(ref_ptr(&domain->address_space));
    return domain->next_address_space;
}

void effdma_domain_transaction_commit(struct effdma_domain *domain)
{
    struct address_space *old_address_space = ref_ptr(&domain->address_space);

    ref_ptr_replace(&domain->address_space, domain->next_address_space);

    address_space_destroy(old_address_space);
}

void effdma_domain_transaction_abort(struct effdma_domain *domain)
{
    address_space_destroy(domain->next_address_space);
}

int effdma_domain_translate(struct effdma_domain *domain, uint64_t siova, uint64_t *tiova)
{
    int rc;
    struct address_space *address_space;

    address_space = ref_ptr_get(&domain->address_space);
    rc = address_space_translate(address_space, siova, tiova);
    ref_ptr_put(&domain->address_space);

    return rc;
}

uint64_t effdma_normalize_size(uint64_t size, enum nvme_cmd_vs_iova_mgmt_szu szu)
{
    switch (szu) {
    case NVME_IOVA_SZU_4K:
        return size * 0x1000;
    case NVME_IOVA_SZU_64K:
        return size * 0x10000;
    case NVME_IOVA_SZU_1M:
        return size * 0x100000;
    case NVME_IOVA_SZU_16M:
        return size * 0x1000000;
    case NVME_IOVA_SZU_256M:
        return size * 0x10000000;
    case NVME_IOVA_SZU_1G:
        return size * 0x40000000;
    case NVME_IOVA_SZU_4G:
        return size * 0x100000000;
    default:
        return 0;
    }
}
