/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdint.h>
#include <string.h>
#include <sys/utsname.h>
#include <infiniband/verbs.h>
#include "nvme_emu_log.h"
#include "compiler.h"
#include "nvme.h"
#include "firmware.h"

#ifdef HAVE_SPDK
#include <spdk/version.h>
#endif

int fw_get_slot_filename(size_t slot_no, char *buf, size_t size)
{
    int rc;

    rc = snprintf(buf, size, "%s/firmware_slot%zu.bin",
            getenv(NVME_FW_UPDATE_PERSISTENT_LOCATION),
            slot_no);

    if (rc >= 0 && rc < size)
        return 0;

    nvmx_error("Failed to get firmware slot file name for slot %zu", slot_no);
    return -1;
}

nvme_json_t *fw_get_bfb_info(nvme_ctrl_t *ctrl, nic_utils_t *ctx, int slot_no)
{
    int rc, fd;
    char *bfb_info;
    char filename[PATH_MAX];

    /* return the cached info if exist */
    if (ctrl->bfb_info[slot_no]) {
        bfb_info = ctrl->bfb_info[slot_no];
        goto out;
    }

    rc = fw_get_slot_filename(slot_no, filename, sizeof(filename));
    if (rc)
        return NULL;

    /* make sure image file is exist */
    if ((fd = open(filename, O_RDONLY)) < 0) {
        nvmx_info("slot %d have no image", slot_no);
        return NULL;
    }
    close(fd);

    bfb_info = nic_utils_exec(ctx, "fw_get_bfb_info", "s", filename);
    if (!bfb_info) {
        nvmx_error("nic_utils exec 'fw_get_bfb_info' failed for slot %d", slot_no);
        return NULL;
    }

    /* fill the cache */
    ctrl->bfb_info[slot_no] = bfb_info;

out:
    return nvme_str_to_json(bfb_info);
}

int fw_get_version_in_bfb(nvme_ctrl_t *ctrl, uint8_t *fw_ver_bfb, int fw_ver_len)
{
    int slot_no;
    int active_slot_no;
    nic_utils_t *ctx;
    nvme_json_t *json;
    bool success, active;

    active_slot_no = -1;

    ctx = nic_utils_open();
    if (!ctx) {
        nvmx_error("Failed to open Nic utils");
        goto out;
    }

    for (slot_no = 1; slot_no <= NVME_FW_SLOTS_COUNT; slot_no++) {
        json = fw_get_bfb_info(ctrl, ctx, slot_no);
        if (!json)
            continue;

        if (nvme_json_get_value(json, "success", &success, sizeof(success))) {
            nvmx_warn("Failed to retrieve success flag for slot:%d", slot_no);
            goto next;
        }

        if (!success)
            goto next;

        if (nvme_json_get_value(json, "active", &active, sizeof(active))) {
            nvmx_warn("Failed to retrieve acrive flag for slot:%d", slot_no);
            goto next;
        }

        if (!active)
            goto next;

        if (nvme_json_get_value(json, "fw", fw_ver_bfb, fw_ver_len)) {
            nvmx_error("Failed to get the fw version");
            nvme_json_free(json);
            break;
        }

        active_slot_no = slot_no;
        nvme_json_free(json);
        break;
next:
        nvme_json_free(json);
    }

    nic_utils_close(ctx);

    if (active_slot_no < 0)
        nvmx_warn("No slot image is in active");
out:
   return active_slot_no;
}

int fw_get_version_in_running(uint8_t *fw_ver_running, int fw_ver_len)
{
    int rc, num_devices;
    struct ibv_device **dev_list;
    struct ibv_context *ctx;
    struct ibv_device_attr_ex device_attr;

    rc = -1;
    dev_list = ibv_get_device_list(&num_devices);
    if (!dev_list || num_devices <= 0) {
        nvmx_error("ibv_get_device_list() failed, errno:%d", errno);
        return rc;
    }

    ctx = ibv_open_device(dev_list[0]);
    if (!ctx) {
        nvmx_error("ibv_open_device() failed, errno:%d", errno);
        goto free_dev_list;
    }

    if (!ibv_query_device_ex(ctx, NULL, &device_attr)) {
        memcpy(fw_ver_running, device_attr.orig_attr.fw_ver,
            nvmx_min(fw_ver_len, strlen(device_attr.orig_attr.fw_ver)));
        rc = 0;
    }

    (void)ibv_close_device(ctx);

free_dev_list:
    ibv_free_device_list(dev_list);

     return rc;
}

void fw_bf_pkg_info_init(nvme_ctrl_t *ctrl, nvme_bf_pkg_info_log_t *bf_pkg_info)
{
    static const char field_name[] = "PRETTY_NAME=\"";
    struct utsname buffer;
    const char *ptr, *end;
    char line[128];
    int rc;

    // Init output with spaces
    memset(bf_pkg_info, ' ', sizeof(nvme_bf_pkg_info_log_t));

    // Get kernel version
    if (uname(&buffer) == 0) {
        memcpy(&bf_pkg_info->krnl, buffer.release,
                nvmx_min(sizeof(bf_pkg_info->krnl), strlen(buffer.release)));
    } else
        nvmx_warn("Failed to get kernel version");

    // Get current running HCA firmware version
    rc = fw_get_version_in_running(bf_pkg_info->nfw, 64);
    if (rc)
        nvmx_warn("Failed to get current running NIC FW version");

    // Get HCA firmware version in running BFB image
    rc = fw_get_version_in_bfb(ctrl, bf_pkg_info->bnfw, 64);
    if (rc < 0)
        nvmx_error("Failed to get NIC FW version in running BFB image");

#ifdef HAVE_SPDK
    // Get SPDK version
    size_t len = snprintf((char *)bf_pkg_info->spdk, sizeof(bf_pkg_info->spdk), "%d.%d.%d",
            SPDK_VERSION_MAJOR, SPDK_VERSION_MINOR, SPDK_VERSION_PATCH);
    if (len < sizeof(bf_pkg_info->spdk))
        bf_pkg_info->spdk[len] = ' ';
#endif

    // Get libsnap version
    memcpy(&bf_pkg_info->lsnap, PACKAGE_VERSION,
            nvmx_min(sizeof(bf_pkg_info->lsnap), strlen(PACKAGE_VERSION)));

    // Get OS name
    FILE *f = fopen("/etc/os-release", "r");
    if (f) {
        while (fgets(line, sizeof(line), f) != NULL)  {
            if (memcmp(line, field_name, sizeof(field_name) - 1) == 0) {
                ptr = line + sizeof(field_name) - 1;
                end = strrchr(ptr, '"');
                if (end)
                    memcpy(bf_pkg_info->os, ptr, nvmx_min(sizeof(bf_pkg_info->os), end - ptr));
                break;
            }
        }
        fclose(f);
    } else
        nvmx_warn("Failed to get OS name");
}
