/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>

#include "config.h"
#include "nvme_emu.h"
#include "nvme_prp.h"
#include "compiler.h"
#include "utils.h"
#include "snap_conf.h"

static const char *q_type2str(int queue_type)
{
    switch (queue_type) {
        case NVME_EMU_QUEUE_SQE:
            return "SQE_ONLY";
        case NVME_EMU_QUEUE_NVME2NVMF_LOCAL:
            return "NVME2NVMF";
        case NVME_EMU_QUEUE_NVME2NVMF_REMOTE:
            return "NVME2NVME+RTARGET";
        default:
            return "UNKNOWN";
    }
}

static int check_q_type(int queue_type)
{
    switch (queue_type) {
        case NVME_EMU_QUEUE_SQE:
        case NVME_EMU_QUEUE_NVME2NVMF_LOCAL:
        case NVME_EMU_QUEUE_NVME2NVMF_REMOTE:
            return NVMX_OK;
        default:
            nvmx_error("unsupported queue type %d", queue_type);
            return NVMX_ERR_NOTSUPPORTED;
    }
}

int nvme_emu_queue_type_add(nvme_emulator_t *dev, int queue_type, bool is_adm,
                            nvme_queue_ops_t *ops)
{
    if (check_q_type(queue_type) != NVMX_OK) 
        return NVMX_ERR;

    ops->q_type = queue_type;

    if (is_adm) 
        dev->admin_q_ops = ops;
    else 
        dev->io_q_ops = ops;

    nvmx_info("using %s layer for %s queue(s)", q_type2str(queue_type),
               is_adm ? "ADMIN" : "IO");

    return NVMX_OK;
}

void nvme_emu_queue_type_del(nvme_emulator_t *dev, int queue_type, bool is_adm)
{
    if (dev->admin_q_ops && dev->admin_q_ops->q_type == queue_type) {
        dev->admin_q_ops = NULL;
    } else if (dev->io_q_ops && dev->io_q_ops->q_type == queue_type) {
       dev->io_q_ops = NULL;
    } else {
        nvmx_warn("queue type %s not found", q_type2str(queue_type));
    }
}

void nvme_emu_set_num_queues(nvme_query_attr_t *dev_caps,
                             struct snap_conf *config,
                             uint32_t *num_queues)
{
    uint32_t nr_io_queues = 1;
    int max_dev_nr_queues;

    max_dev_nr_queues = nvmx_min(dev_caps->max_emulated_sq_num,
                                 dev_caps->max_emulated_cq_num);
    if (!config)
        goto out;

    nr_io_queues = nvmx_min(snap_config_get_num_queues(config),
                            max_dev_nr_queues - 1);


out:
    *num_queues = 1 + nr_io_queues; // +1 for admin
}

void nvme_emu_qops_clone(nvme_emulator_t *dev,
                                nvme_queue_ops_t *ops, int qid)
{
    if (qid)
        memcpy(ops, dev->io_q_ops, sizeof(*ops));
    else
        memcpy(ops, dev->admin_q_ops, sizeof(*ops));

    ops->qid = qid;
    ops->dev = dev;
}

