/*-
 *   BSD LICENSE
 *
 *   Copyright (c) Intel Corporation.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * NVMe over Fabrics transport-independent functions - based on SPDK
 */

#include "config.h"
#include "nvmf_emu_ctrl.h"
#include "nvmf.h"

int nvme_fabric_qpair_connect(struct nvmf_emu_ctrl *ctrl, uint16_t qid,
                              uint32_t num_entries)
{
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvmf_fabric_connect_cmd *cmd;
    nvmf_fabric_connect_rsp *rsp;
    nvmf_fabric_connect_data *nvmf_data;
    int rc;

    if (num_entries == 0 || num_entries > ctrl->max_io_queue_size) {
        nvmx_error("queue %d: num_entries %d is invalid. Max %d",
                   qid, num_entries, ctrl->max_io_queue_size);
        return -EINVAL;
    }

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, qid, true);
    if (!nvmf_req) {
        nvmx_error("queue %d: out of resources", qid);
        return -EAGAIN;
    }

    cmd = (nvmf_fabric_connect_cmd *)&nvmf_req->cmd;
    nvmf_data = (nvmf_fabric_connect_data *)nvmf_req->payload;

    memset(cmd, 0, sizeof(*cmd));
    cmd->opcode = NVME_OPCODE_FABRIC;
    cmd->fctype = NVMF_FABRIC_COMMAND_CONNECT;
    cmd->qid = htole16(qid);
    cmd->sqsize = htole16(num_entries - 1);
    cmd->kato = htole32(ctrl->ka_timeout_ms);

    if (qid == 0)
        nvmf_data->cntlid = htole16(0xFFFF);
    else
        nvmf_data->cntlid = htole16(ctrl->cntlid);

    uuid_copy(nvmf_data->hostid, ctrl->hostid);
    snprintf((char *)nvmf_data->hostnqn, sizeof(nvmf_data->hostnqn), "%s", ctrl->hostnqn);
    snprintf((char *)nvmf_data->subnqn, sizeof(nvmf_data->subnqn), "%s", ctrl->id.subnqn);

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, qid, nvmf_req, nvmf_data,
                             sizeof(*nvmf_data), &status);
    if (rc || NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("queue %d: Connect command failed. rc %d status %x", qid,
                   rc, status.cqe.status);
        ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);
        return rc ? rc : -1;
    }

    if (!qid) {
        rsp = (nvmf_fabric_connect_rsp *)&status.cqe;
        ctrl->cntlid = rsp->status_code_specific.success.cntlid;
        nvmx_debug("CNTLID %d", ctrl->cntlid);
    }

    ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);

    return 0;
}

int nvme_fabric_prop_get_cmd(struct nvmf_emu_ctrl *ctrl, uint32_t offset,
                             uint8_t size, void *value)
{
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvmf_fabric_prop_get_cmd *cmd;
    nvmf_fabric_prop_get_rsp *response;
    uint32_t *u32_val;
    uint64_t *u64_val;
    int rc;

    if (size != NVMF_PROP_SIZE_4 && size != NVMF_PROP_SIZE_8)
        return -EINVAL;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvmf_fabric_prop_get_cmd *)&nvmf_req->cmd;
    cmd->opcode = NVME_OPCODE_FABRIC;
    cmd->fctype = NVMF_FABRIC_COMMAND_PROPERTY_GET;
    cmd->ofst = offset;
    cmd->attrib.size = size;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, NULL, 0, &status);
    if (rc) {
        nvmx_error("Property Get failed. rc %d status 0x%x", rc,
                   NVME_CQE_STATUS_IS_ERROR(status.cqe.status));
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
        return rc;
    }

    /*
     * We must provide the response to the user even though
     * the CQE status is not success
     */
    response = (nvmf_fabric_prop_get_rsp *)&status.cqe;
    if (size == NVMF_PROP_SIZE_4) {
        u32_val  = value;
        *u32_val = response->value.u32.low;
    }
    else {
        u64_val  = value;
        *u64_val = response->value.u64;
    }

    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);

    return NVME_CQE_STATUS_IS_ERROR(status.cqe.status);
}

int nvme_fabric_prop_set_cmd(struct nvmf_emu_ctrl *ctrl, uint32_t offset,
                             uint8_t size, uint64_t value)
{
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvmf_fabric_prop_set_cmd *cmd;
    int rc;

    if (size != NVMF_PROP_SIZE_4 && size != NVMF_PROP_SIZE_8)
        return -EINVAL;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvmf_fabric_prop_set_cmd *)&nvmf_req->cmd;
    cmd->opcode = NVME_OPCODE_FABRIC;
    cmd->fctype = NVMF_FABRIC_COMMAND_PROPERTY_SET;
    cmd->ofst = offset;
    cmd->attrib.size = size;
    cmd->value = value;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, NULL, 0, &status);
    if (rc) {
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
        return rc;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Property Set failed %u", status.cqe.status);
        rc = -1;
    }

    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);

    return rc;
}

/* This command is also part of Non-fabric cmd set */
int nvme_ctrl_set_feature_cmd(struct nvmf_emu_ctrl *ctrl, uint8_t feature,
                              uint32_t dw11, uint32_t *res)
{

    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_set_feature_t *cmd;
    int rc;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvme_cmd_set_feature_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_SET_FEATURES;
    cmd->fid = feature;
    cmd->dw11 = dw11;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, NULL, 0, &status);
    if (rc) {
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
        return rc;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Set features failed %u", status.cqe.status);
        rc = -1;
    } else {
        *res = le32toh(status.cqe.result);
    }

    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);

    return rc;
}

/* This command is also part of Non-fabric cmd set */
int nvme_ctrl_get_feature_cmd(struct nvmf_emu_ctrl *ctrl, uint8_t feature,
                              uint32_t *res)
{

    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_get_feature_t *cmd;
    int rc;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvme_cmd_get_feature_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_GET_FEATURES;
    cmd->fid = feature;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, NULL, 0, &status);
    if (rc) {
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
        return rc;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Get features failed %u feature %u", status.cqe.status,
                   feature);
        rc = -1;
    } else {
        *res = le32toh(status.cqe.result);
    }

    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);

    return rc;
}

int nvme_ctrl_get_log_page_cmd(struct nvmf_emu_ctrl *ctrl, nvme_log_page_t *log,
                               uint32_t nsid, uint8_t page_id, uint8_t lsp)
{
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_get_page_log_t *cmd;
    int rc;
    uint16_t log_size;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvme_cmd_get_page_log_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_GET_LOG_PAGE;
    cmd->common.nsid = htole32(nsid);
    cmd->lid = page_id;
    cmd->lsp = lsp & 0xF;

    switch (page_id) {
    case NVME_LOG_ERROR_INFO:
        log_size = sizeof(log->error_log);
        break;
    case NVME_LOG_SMART_INFO:
        log_size = sizeof(log->smart_log);
        break;
    case NVME_LOG_FW_SLOT_INFO:
        log_size = sizeof(log->fw_slot_info_log);
        break;
    case NVME_LOG_CHANGED_NS_LIST_INFO:
        log_size = sizeof(log->changed_nslist_log);
        break;
    default:
        nvmx_error("Unsupported log page id %d requested", page_id);
        rc = -EINVAL;
        goto out;
    }

    /* We currently do not support extended Get Log Pager interface */
    cmd->numdl = htole16(log_size / 4 - 1);

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, nvmf_req->payload,
                                     log_size, &status);
    if (rc) {
        nvmx_error("Failed to send Get Log Page %d", page_id);
        goto out;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Get Log Page failed %u", status.cqe.status);
        rc = -1;
    } else {
        memcpy(log, nvmf_req->payload, log_size);
    }

out:
    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
    return rc;
}

int nvme_identify_cmd(struct nvmf_emu_ctrl *ctrl, void *payload,
                      int payload_size, uint32_t nsid, uint8_t cns)
{
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_identify_t *cmd;
    int rc;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, true);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvme_cmd_identify_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_IDENTIFY;
    cmd->cns = cns;
    cmd->common.nsid = htole32(nsid);

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, 0, nvmf_req, nvmf_req->payload,
                             payload_size, &status);
    if (rc) {
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
        return rc;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Identify %u list failed %u", cns, status.cqe.status);
        rc = -1;
    } else {
        memcpy(payload, nvmf_req->payload, payload_size);
    }

    ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);

    return rc;
}

int nvme_ctrl_submit_async_event_cmd(struct nvmf_emu_ctrl *ctrl)
{
    nvmf_req_t *nvmf_req;
    nvme_cmd_t *cmd;
    int rc;

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, false);
    if (!nvmf_req)
        return -EAGAIN;

    cmd = (nvme_cmd_t *)&nvmf_req->cmd;
    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_ASYNC_EV_REQ;

    rc = ctrl->cmd_ops.send_cmd_async(ctrl, 0, nvmf_req, NULL, 0);
    if (rc) {
        nvmx_error("Failed to send Async event request");
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
    }

    return rc;
}
