#include "config.h"
#include "nvme_emu_ep.h"
#include "nvme_emu_io_driver.h"
#include "nvme_posix_ctrl.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "snap_conf.h"

static int nvme_emu_posix_write_sgl(nvme_emu_ep_t *ep,
                                    nvme_io_driver_req_t *req,
                                    const nvme_cmd_rw_t *rw)
{
    nvme_posix_ctrl_t *ctrl = (nvme_posix_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    int ret;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    nvme_sgl_read(ops, sgl, ctrl->io_bbuf, count);

    ret = pwrite(ctrl->io_fd, ctrl->io_bbuf, count, offset);
    if (ret < 0) {
        nvmx_error("disk write sgl failed: slb %lu nlb %u sgl: 0x%p pbits %u",
                   slba, nlb, sgl, ops->page_bits);
        return NVMX_ERR;
    }
    return NVMX_OK;
}

static int nvme_emu_posix_read_sgl(nvme_emu_ep_t *ep,
                                   nvme_io_driver_req_t *req,
                                   const nvme_cmd_rw_t *rw)
{
    nvme_posix_ctrl_t *ctrl = (nvme_posix_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    int ret;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    ret = pread(ctrl->io_fd, ctrl->io_bbuf, count, offset);
    if (ret < 0) {
        nvmx_error("disk read sgl failed: slb %lu nlb %u sgl: 0x%p pbits %u",
                   slba, nlb, sgl, ops->page_bits);
        return NVMX_ERR;
    }

    nvme_sgl_write(ops, sgl, ctrl->io_bbuf, count);

    return NVMX_OK;
}


static int nvme_emu_posix_write_prp(nvme_emu_ep_t *ep,
                                    nvme_io_driver_req_t *req,
                                    const nvme_cmd_rw_t *rw)
{
    nvme_posix_ctrl_t *ctrl = (nvme_posix_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    int ret;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    nvme_prp_read_short(ops, prp1, prp2, ctrl->io_bbuf, count);

    ret = pwrite(ctrl->io_fd, ctrl->io_bbuf, count, offset);
    if (ret < 0) {
        nvmx_error("disk write failed (%m): slb %lu nlb %u prp1: 0x%llx prp2: 0x%llx pbits %u",
                   slba, nlb,
                   (unsigned long long)prp1,
                   (unsigned long long)prp2, ops->page_bits);
        return NVMX_ERR;
    }
    return NVMX_OK;
}

static int nvme_emu_posix_read_prp(nvme_emu_ep_t *ep,
                                   nvme_io_driver_req_t *req,
                                   const nvme_cmd_rw_t *rw)
{
    nvme_posix_ctrl_t *ctrl = (nvme_posix_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    int ret;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    ret = pread(ctrl->io_fd, ctrl->io_bbuf, count, offset);
    if (ret < 0) {
        nvmx_error("disk read failed (%m): slb %lu nlb %u prp1: 0x%llx prp2: 0x%llx pbits %u",
                   slba, nlb,
                   (unsigned long long)prp1,
                   (unsigned long long)prp2, ops->page_bits);
        return NVMX_ERR;
    }
    nvme_prp_write_short(ops, prp1, prp2, ctrl->io_bbuf, count);
    return NVMX_OK;
}

static int nvme_emu_posix_create_cq(nvme_emu_ep_t *ep, uint16_t cqid,
                                   uint16_t size)
{
    return 0;
}

static void nvme_emu_posix_delete_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static void nvme_emu_posix_deactivate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static int nvme_emu_posix_activate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    return 0;
}

static int nvme_emu_posix_create_sq(nvme_emu_ep_t *ep, uint16_t sqid,
                                   uint16_t size, uint16_t cqid)
{
    return 0;
}

static void nvme_emu_posix_delete_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
}

static void nvme_emu_posix_deactivate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
}

static int nvme_emu_posix_activate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    return 0;
}

void nvme_emu_delete_posix_ctrl(nvme_posix_ctrl_t *ctrl)
{
    free(ctrl->io_bbuf);
    close(ctrl->io_fd);
    free(ctrl);
}

int nvme_emu_create_posix_ctrl(nvme_emu_ep_t *ep)
{
    nvme_posix_ctrl_t *ctrl;
    snap_conf_t *sconfig = ep->io_driver->sconfig;
    const snap_config_backend_t *backend = snap_config_get_backend(sconfig, 0);

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        return NVMX_ERR;

    ctrl->io_fd = open(ep->name, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
    if (ctrl->io_fd < 0) {
        nvmx_error("failed to open %s", ep->name);
        goto out_err;
    }

    ctrl->disk_size_b = snap_config_get_backend_size_b(backend);
    ctrl->disk_block_order = snap_config_get_backend_block_order(backend);

    ctrl->io_bbuf = calloc(ctrl->disk_size_b, sizeof(*ctrl->io_bbuf));
    if (!ctrl->io_bbuf) {
        nvmx_error("failed to allocate bbuf sized %zd",
                   ctrl->disk_size_b);
        goto close_file;
    }

    if ((lseek(ctrl->io_fd, ctrl->disk_size_b, SEEK_SET) == ((off_t)-1)) ||
         lseek(ctrl->io_fd, 0, SEEK_SET) == ((off_t)-1)) {
        nvmx_error("Failed to lseek io_fd %d: %m", ctrl->io_fd);
        goto free_io_bbuf;
    }

    ep->ctrl = ctrl;
    ep->ep_ops.read_prp  = nvme_emu_posix_read_prp;
    ep->ep_ops.write_prp = nvme_emu_posix_write_prp;
    ep->ep_ops.read_sgl  = nvme_emu_posix_read_sgl;
    ep->ep_ops.write_sgl = nvme_emu_posix_write_sgl;
    ep->ep_ops.create_cq = nvme_emu_posix_create_cq;
    ep->ep_ops.delete_cq = nvme_emu_posix_delete_cq;
    ep->ep_ops.activate_cq = nvme_emu_posix_activate_cq;
    ep->ep_ops.deactivate_cq = nvme_emu_posix_deactivate_cq;
    ep->ep_ops.create_sq = nvme_emu_posix_create_sq;
    ep->ep_ops.delete_sq = nvme_emu_posix_delete_sq;
    ep->ep_ops.activate_sq = nvme_emu_posix_activate_sq;
    ep->ep_ops.deactivate_sq = nvme_emu_posix_deactivate_sq;

    ctrl->ep = ep;

    return NVMX_OK;

free_io_bbuf:
    free(ctrl->io_bbuf);
close_file:
    close(ctrl->io_fd);
out_err:
    free(ctrl);
    return NVMX_ERR;
}

int nvme_start_posix_ctrl(nvme_emu_ep_t *ep)
{
    nvme_posix_ctrl_t *ctrl = ep->ctrl;

    /* Build namespace, if needed */
    if (ctrl->disk_size_b > 0) {
        nvme_id_ns_descriptor_t ns_desc[2];

        ns_desc[0].nidt = NVME_NS_IDENTIFIER_UUID;
        ns_desc[0].nidl = 0x10;
        memset(&ns_desc[0].nid.uuid, 0, ns_desc[0].nidl);
        memset(&ns_desc[1], 0, sizeof(ns_desc[1]));
        nvme_emu_ep_attach_volume(ep, 1, ctrl->disk_size_b,
                                  ctrl->disk_block_order, 0, ns_desc, ctrl);
    }

    return NVMX_OK;
}

void nvme_stop_posix_ctrl(nvme_emu_ep_t *ep)
{
    nvme_emu_ep_detach_volume(ep, 1);
}
