#include <time.h>
#include "config.h"
#include "nvmf_emu_ctrl.h"
#include "nvme_emu_io_driver.h"
#include "nvmf_rdma.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "json/nvme_json.h"
#include "utils.h"
#include "snap_conf.h"

/* NVMf controller can support metadata only if target can
 * support any data size (mdts = 0) or strictly larger than
 * the data size the host may send
 */
#define NVMF_CTRL_IS_MD_CONFIGURABLE(tmdts, hmdts) ((tmdts == 0) || \
                                                   ((hmdts != 0) && tmdts > hmdts))

/* Declarations */
static nvme_ns_t *nvmf_ctrl_find_namespace_lock(nvmf_emu_ctrl_t *ctrl, uint32_t nsid);
static int nvmf_emu_enable_ctrl(nvmf_emu_ctrl_t *ctrl);
static void nvmf_emu_disable_ctrl(nvmf_emu_ctrl_t *ctrl);
static int nvmf_set_num_io_queues(nvmf_emu_ctrl_t *ctrl);
static void nvmf_ctrl_clear_ns_list(nvmf_emu_ctrl_t *ctrl);
static int nvmf_ctrl_verify_ctrl_configuration(nvmf_emu_ctrl_t *ctrl,
                                               snap_conf_t *sconfig);
static int nvmf_ctrl_configure_async_events(nvmf_emu_ctrl_t *ctrl);
static void nvmf_ctrl_deconfigure_async_events(nvmf_emu_ctrl_t *ctrl);
static void nvmf_ctrl_update_namespaces(nvmf_emu_ctrl_t *ctrl);
static void nvmf_ctrl_update_changed_namespaces(nvmf_emu_ctrl_t *ctrl,
                                                uint32_t *ns_list,
                                                size_t ns_list_size);

static int nvmf_emu_ctrl_verify_rw_cmd(const nvme_cmd_rw_t *rw, nvme_ns_t *ns)
{
    if ((le64_to_cpu(rw->slba) + le16_to_cpu(rw->nlb) + 1) << ns->block_order >
        ns->size_b) {
        nvmx_error("Write address outside of disk boundaries");
        return -EINVAL;
    }

    return 0;
}

static int nvmf_emu_ctrl_write_prp(nvme_emu_ep_t *ep,
                                   nvme_io_driver_req_t *req,
                                   const nvme_cmd_rw_t *rw)
{
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_rw_t *cmd;
    nvme_ns_t *ns;
    nvme_queue_ops_t *ops = req->async_req.ops;
    int rc, qid = ops->qid;
    size_t count;

    ns = nvmf_ctrl_find_namespace_lock(ctrl, le32_to_cpu(rw->common.nsid));
    if (!ns)
        return NVMX_ERR;

    rc = nvmf_emu_ctrl_verify_rw_cmd(rw, ns);
    if (rc)
        return NVMX_ERR;

    /* Currently traffic is sync on qid 1 */
    nvmf_req = ctrl->cmd_ops.get_request(ctrl, qid, true);
    if (!nvmf_req)
        return NVMX_ERR;

    cmd = (nvme_cmd_rw_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_IO_WRITE;
    cmd->common.nsid = rw->common.nsid;
    cmd->slba = rw->slba;
    cmd->nlb = rw->nlb; // zero based ??
    cmd->control = cpu_to_le16(0);
    cmd->dsmgmt = cpu_to_le32(0);

    count = (size_t)nlb << ns->block_order;
    req->async_req.len = count;

    nvme_prp_read_short(ops, prp1, prp2, nvmf_req->payload, count);

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, qid, nvmf_req, nvmf_req->payload,
                             count, &status);
    if (rc) {
        nvmx_error("write command failed for queue %d", qid);
        ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);
        return NVMX_ERR;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Write bad rsp %u", status.cqe.status);
        rc = NVMX_ERR;
    } else {
        rc = NVMX_OK;
    }

    ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);

    return rc;
}

static int nvmf_emu_ctrl_read_prp(nvme_emu_ep_t *ep,
                                  nvme_io_driver_req_t *req,
                                  const nvme_cmd_rw_t *rw)
{
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    nvmf_completion_status_t status = {};
    nvmf_req_t *nvmf_req;
    nvme_cmd_rw_t *cmd;
    nvme_ns_t *ns;
    nvme_queue_ops_t *ops = req->async_req.ops;
    int rc, qid = ops->qid;
    size_t count;

    ns = nvmf_ctrl_find_namespace_lock(ctrl, le32_to_cpu(rw->common.nsid));
    if (!ns)
        return NVMX_ERR;

    rc = nvmf_emu_ctrl_verify_rw_cmd(rw, ns);
    if (rc)
        return NVMX_ERR;

    /* Currently traffic is sync on qid 1 */
    nvmf_req = ctrl->cmd_ops.get_request(ctrl, qid, true);
    if (!nvmf_req)
        return NVMX_ERR;

    cmd = (nvme_cmd_rw_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_IO_READ;
    cmd->common.nsid = rw->common.nsid;
    cmd->slba = rw->slba;
    cmd->nlb = rw->nlb; // zero based ??
    cmd->control = cpu_to_le16(0);
    cmd->dsmgmt = cpu_to_le32(0);

    count = (size_t)nlb << ns->block_order;
    req->async_req.len = count;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, qid, nvmf_req, nvmf_req->payload,
                             count, &status);
    if (rc) {
        nvmx_error("Read command failed for queue %d", qid);
        ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);
        return NVMX_ERR;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Read bad rsp %u", status.cqe.status);
        rc = NVMX_ERR;
    } else {
        nvme_prp_write_short(ops, prp1, prp2, nvmf_req->payload, count);
        rc = NVMX_OK;
    }

    ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);

    return rc;
}

static int nvme_emu_ctrl_write_sgl(nvme_emu_ep_t *ep,
                                   nvme_io_driver_req_t *req,
                                   const nvme_cmd_rw_t *rw)
{
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    nvmf_completion_status_t status = {};
    nvme_queue_ops_t *ops = req->async_req.ops;
    int qid = ops->qid;
    nvmf_req_t *nvmf_req;
    nvme_cmd_rw_t *cmd;
    nvme_ns_t *ns;
    int rc;
    size_t count;

    ns = nvmf_ctrl_find_namespace_lock(ctrl, le32_to_cpu(rw->common.nsid));
    if (!ns)
        return NVMX_ERR;

    rc = nvmf_emu_ctrl_verify_rw_cmd(rw, ns);
    if (rc)
        return NVMX_ERR;

    /* Currently traffic is sync on qid 1 */
    nvmf_req = ctrl->cmd_ops.get_request(ctrl, qid, true);
    if (!nvmf_req)
        return NVMX_ERR;

    cmd = (nvme_cmd_rw_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_IO_WRITE;
    cmd->common.nsid = rw->common.nsid;
    cmd->slba = rw->slba;
    cmd->nlb = rw->nlb; // zero based ??
    cmd->control = cpu_to_le16(0);
    cmd->dsmgmt = cpu_to_le32(0);

    count = (size_t)nlb << ns->block_order;
    req->async_req.len = count;

    nvme_sgl_read(ops, &rw->common.dptr.sgl, nvmf_req->payload, count);

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, qid, nvmf_req, nvmf_req->payload,
                                     count, &status);
    if (rc) {
        nvmx_error("write command failed for queue %d", qid);
        ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);
        return NVMX_ERR;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Write bad rsp %u", status.cqe.status);
        rc = NVMX_ERR;
    } else {
        rc = NVMX_OK;
    }

    ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);

    return rc;
}

static int nvme_emu_ctrl_read_sgl(nvme_emu_ep_t *ep,
                                  nvme_io_driver_req_t *req,
                                  const nvme_cmd_rw_t *rw)
{
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    nvmf_completion_status_t status = {};
    nvme_queue_ops_t *ops = req->async_req.ops;
    int qid = ops->qid;
    nvmf_req_t *nvmf_req;
    nvme_cmd_rw_t *cmd;
    nvme_ns_t *ns;
    int rc;
    size_t count;

    ns = nvmf_ctrl_find_namespace_lock(ctrl, le32_to_cpu(rw->common.nsid));
    if (!ns)
        return NVMX_ERR;

    rc = nvmf_emu_ctrl_verify_rw_cmd(rw, ns);
    if (rc)
        return NVMX_ERR;

    /* Currently traffic is sync on qid 1 */
    nvmf_req = ctrl->cmd_ops.get_request(ctrl, qid, true);
    if (!nvmf_req)
        return NVMX_ERR;

    cmd = (nvme_cmd_rw_t *)&nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_IO_READ;
    cmd->common.nsid = rw->common.nsid;
    cmd->slba = rw->slba;
    cmd->nlb = rw->nlb; // zero based ??
    cmd->control = cpu_to_le16(0);
    cmd->dsmgmt = cpu_to_le32(0);

    count = (size_t)nlb << ns->block_order;
    req->async_req.len = count;

    rc = ctrl->cmd_ops.send_cmd_sync(ctrl, qid, nvmf_req, nvmf_req->payload,
                                     count, &status);
    if (rc) {
        nvmx_error("Read command failed for queue %d", qid);
        ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);
        return NVMX_ERR;
    }

    if (NVME_CQE_STATUS_IS_ERROR(status.cqe.status)) {
        nvmx_error("Read bad rsp %u", status.cqe.status);
        rc = NVMX_ERR;
    } else {
        nvme_sgl_write(ops, &rw->common.dptr.sgl, nvmf_req->payload, count);
        rc = NVMX_OK;
    }

    ctrl->cmd_ops.put_request(ctrl, qid, nvmf_req);

    return rc;
}

static enum nvmf_trtype nvme_emu_ep_trtype(enum nvme_emu_ep_type ep_type)
{
    if (ep_type == NVME_EMU_EP_TYPE_NVMF_RDMA)
        return NVMF_TRTYPE_RDMA;

    return 0;
}

/*
 * Must be called with ep_lock held
 */
static int nvmf_emu_ctrl_create_cq(nvme_emu_ep_t *ep, uint16_t cqid,
                                   uint16_t size)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;

    if (size > ctrl->max_io_queue_size) {
        nvmx_warn("queue %u: Requested queue size is too big, "
                  "which may cause performance degradation.",
                  cqid, size, ctrl->max_io_queue_size);
    }

    if (cqid >= ctrl->nr_queues) {
        nvmx_error("queue %u: can't create cq. Supported nr_queues is %u",
                   cqid, ctrl->nr_queues);
        return NVMX_ERR;
    }

    return 0;
}

/*
 * Must be called with ep_lock held
 */
static void nvmf_emu_ctrl_delete_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

/*
 * Must be called with ep_lock held
 */
static void nvmf_emu_ctrl_deactivate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

/*
 * Must be called with ep_lock held
 */
static int nvmf_emu_ctrl_activate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    return 0;
}

/*
 * Must be called with ep_lock held
 */
static int nvmf_emu_ctrl_create_sq(nvme_emu_ep_t *ep, uint16_t sqid,
                                   uint16_t size, uint16_t cqid)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    int ret;

    if (size > ctrl->max_io_queue_size) {
        nvmx_warn("queue %u: Requested queue size is too big, "
                  "which may cause performance degradation. "
                  "Reducing it from %u to %u",
                  sqid, size, ctrl->max_io_queue_size);
        size = ctrl->max_io_queue_size;
    }

    if (sqid >= ctrl->nr_queues) {
        nvmx_error("queue %u: can't create sq. Supported nr_queues is %u",
                   sqid, ctrl->nr_queues);
        return NVMX_ERR;
    }

    ret = ctrl->q_ops.create(ctrl, sqid, size);
    if (ret)
        return NVMX_ERR;

    return 0;
}

/*
 * Must be called with ep_lock held
 */
static void nvmf_emu_ctrl_delete_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;

    ctrl->q_ops.destroy(ctrl, sqid);
}

/*
 * Must be called with ep_lock held
 */
static void nvmf_emu_ctrl_deactivate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;

    ctrl->q_ops.deactivate(ctrl, sqid);
}

/*
 * Must be called with ep_lock held
 */
static int nvmf_emu_ctrl_activate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;

    return ctrl->q_ops.activate(ctrl, sqid);
}

static enum nvmf_adrfam nvme_emu_ep_adrfam(enum nvme_emu_ep_fam ep_fam)
{
    if (ep_fam == NVME_EMU_ADRFAM_IPV4)
        return NVMF_ADRFAM_IPV4;
    if (ep_fam == NVME_EMU_ADRFAM_IPV6)
        return NVMF_ADRFAM_IPV6;
    if (ep_fam == NVME_EMU_ADRFAM_IB)
        return NVMF_ADRFAM_IB;
    if (ep_fam == NVME_EMU_ADRFAM_FC)
        return NVMF_ADRFAM_FC;

    return 0;
}

int nvme_emu_ep_to_nvmf_ep(nvme_emu_ep_t *ep, struct nvmf_transport_id *id)
{
    id->trtype = nvme_emu_ep_trtype(ep->ep_type);
    if (!id->trtype)
        return EINVAL;
    id->adrfam = nvme_emu_ep_adrfam(ep->ep_fam);
    if (!id->adrfam)
        return EINVAL;
    memcpy(id->subnqn, ep->name, NVMF_NQN_MAX_LEN);
    memcpy(id->traddr, ep->addr, NVMF_TRADDR_MAX_LEN);
    memcpy(id->trsvcid, ep->port, NVMF_TRSVCID_MAX_LEN);

    return 0;
}

static int nvmf_emu_get_src_addr(nvme_emu_ep_t *ep, char *src_addr, size_t src_addr_sz)
{
    char if_name[IFACE_MAX_LEN] = {0};

    if (ep->io_driver->dev) {
        if (dev_to_iface(snap_config_get_rdma_dev(ep->io_driver->sconfig),
                         if_name)) {
            nvmx_error("Failed to translate HCA %s to network interface",
                       snap_config_get_rdma_dev(ep->io_driver->sconfig));
            return -ENODEV;
	}
    } else
        return 0; /* If no dev was configured, use default src_addr (NULL) */

    switch (ep->ep_fam) {
    case NVME_EMU_ADRFAM_IPV4:
        break;
    default:
        nvmx_error("Address Family %d is not supported", ep->ep_fam);
        return -EINVAL;
    }

    return iface_to_ipaddr(if_name, src_addr, src_addr_sz);
}

static void nvmf_emu_destroy_admin_queue(nvmf_emu_ctrl_t *ctrl)
{
    ctrl->q_ops.deactivate(ctrl, 0);
    ctrl->q_ops.destroy(ctrl, 0);
}

static int nvmf_emu_create_admin_queue(nvmf_emu_ctrl_t *ctrl)
{
    int ret;

    ret = ctrl->q_ops.create(ctrl, 0, NVMF_MIN_ADMIN_QUEUE_ENTRIES);
    if (ret)
        return NVMX_ERR;

    ret = ctrl->q_ops.activate(ctrl, 0);
    if (ret) {
        ret = NVMX_ERR;
        ctrl->q_ops.destroy(ctrl, 0);
    }

    return ret;
}

static void nvmf_emu_set_keep_alive_tmo(struct snap_conf *sconfig,
                                        nvmf_emu_ctrl_t *ctrl)
{
    const snap_config_be_path_t *path;

    path = snap_config_get_path(sconfig, 0, 0);
    ctrl->ka_timeout_ms = path->ka_timeout_ms;

    if (ctrl->ka_timeout_ms < NVMF_MIN_KA_TIMEOUT_MS) {
        nvmx_warn("Wrong keep-alive timeout value: ka_timeout=%d, "
                   "it cannot be less than %d ms",
                   ctrl->ka_timeout_ms, NVMF_MIN_KA_TIMEOUT_MS);
        ctrl->ka_timeout_ms = NVMF_MIN_KA_TIMEOUT_MS;
    }
}

static void nvmf_emu_free_ctrl(nvmf_emu_ctrl_t *ctrl, enum nvmf_trtype trtype)
{
    switch (trtype) {
    case NVMF_TRTYPE_RDMA:
        nvmf_rdma_free_ctrl(ctrl);
        break;
    default:
        nvmx_error("Currently destroying %d type is not supported", trtype);
        break;
    }
}

static nvmf_emu_ctrl_t *nvmf_emu_alloc_ctrl(nvme_emu_ep_t *ep)
{
    nvmf_emu_ctrl_t *ctrl;

    switch (nvme_emu_ep_trtype(ep->ep_type)) {
    case NVMF_TRTYPE_RDMA:
        ctrl = nvmf_rdma_alloc_ctrl(ep);
        break;
    default:
        nvmx_error("Currently creating %d type is not supported",
                   nvme_emu_ep_trtype(ep->ep_type));
        ctrl = NULL;
    }

    return ctrl;
}

void nvmf_emu_stop_ctrl(nvme_emu_ep_t *ep, bool clear_namespaces)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;

    /*
     * We don't clear ns list here, as it is done via
     * asynchronous events mechanism
     */
    nvmf_ctrl_deconfigure_async_events(ctrl);
    if (clear_namespaces) {
        pthread_mutex_lock(&ctrl->ns_list_lock);
        nvmf_ctrl_clear_ns_list(ctrl);
        pthread_mutex_unlock(&ctrl->ns_list_lock);
    }
    memset(&ctrl->cdata, 0, sizeof(ctrl->cdata));
    nvmf_emu_disable_ctrl(ctrl);
    nvmf_emu_destroy_admin_queue(ctrl);
    sem_destroy(&ctrl->fail_sem);
}

int nvmf_emu_start_ctrl(nvme_emu_ep_t *ep)
{
    nvmf_emu_ctrl_t *ctrl = ep->ctrl;
    int err = 0;

    err = sem_init(&ctrl->fail_sem, 0, 0);
    if (err)
        goto out_err;

    err = nvmf_emu_create_admin_queue(ctrl);
    if (err)
        goto destroy_sem;

    err = nvmf_emu_enable_ctrl(ctrl);
    if (err)
        goto destroy_admin_queue;

    err = nvmf_identify_ctrl(ctrl);
    if (err)
        goto disable_ctrl;

    err = nvmf_set_num_io_queues(ctrl);
    if (err)
        goto disable_ctrl;

    nvmf_ctrl_update_namespaces(ctrl);

    err = nvmf_ctrl_configure_async_events(ctrl);
    if (err)
        goto disable_ctrl;

    err = nvmf_ctrl_verify_ctrl_configuration(ctrl, ep->io_driver->sconfig);
    if (err)
        goto deconfigure_aens;

    sem_post(&ctrl->fail_sem);
    nvmx_info("NVMf ctrl %u started on (%s:%s) %s type %d src addr (%s)",
              ctrl->cntlid, ep->addr, ep->port, ep->name, ctrl->id.trtype,
              ctrl->src_addr[0] ? ctrl->src_addr : "unknown");

    return 0;

deconfigure_aens:
    nvmf_ctrl_deconfigure_async_events(ctrl);
disable_ctrl:
    nvmf_emu_disable_ctrl(ctrl);
destroy_admin_queue:
    nvmf_emu_destroy_admin_queue(ctrl);
destroy_sem:
    sem_destroy(&ctrl->fail_sem);
out_err:
    return err;
}

int nvmf_emu_create_ctrl(nvme_emu_ep_t *ep)
{
    nvmf_emu_ctrl_t *ctrl;
    const snap_config_backend_t *backend;
    int err;

    if (strcmp(ep->name, NVMF_DISCOVERY_NQN) == 0) {
        /* Currently discovery service is not supported*/
        return -EINVAL;
    }

    ctrl = nvmf_emu_alloc_ctrl(ep);
    if (!ctrl)
        return -EAGAIN;

    err = nvmf_emu_get_src_addr(ep, ctrl->src_addr, INET_ADDRSTRLEN);
    if (err) {
        err = -EINVAL;
        goto free_ctrl;
    }

    err = nvme_emu_ep_to_nvmf_ep(ep, &ctrl->id);
    if (err)
        goto free_ctrl;

    err = pthread_mutex_init(&ctrl->state_lock, NULL);
    if (err)
        goto free_ctrl;

    err = pthread_mutex_init(&ctrl->ns_list_lock, NULL);
    if (err)
        goto free_state_lock;

    TAILQ_INIT(&ctrl->ns_list);

    backend = snap_config_get_backend(ep->io_driver->sconfig,
                                      ep->io_driver->index);
    snprintf(ctrl->hostnqn, sizeof(ctrl->hostnqn), "%s", ep->nqn);
    snap_config_get_backend_hostid(backend, ctrl->hostid);
    nvmf_emu_set_keep_alive_tmo(ep->io_driver->sconfig, ctrl);

    ep->ctrl = ctrl;
    ep->ep_ops.read_prp = nvmf_emu_ctrl_read_prp;
    ep->ep_ops.write_prp = nvmf_emu_ctrl_write_prp;
    ep->ep_ops.read_sgl  = nvme_emu_ctrl_read_sgl;
    ep->ep_ops.write_sgl = nvme_emu_ctrl_write_sgl;
    ep->ep_ops.create_cq = nvmf_emu_ctrl_create_cq;
    ep->ep_ops.delete_cq = nvmf_emu_ctrl_delete_cq;
    ep->ep_ops.activate_cq = nvmf_emu_ctrl_activate_cq;
    ep->ep_ops.deactivate_cq = nvmf_emu_ctrl_deactivate_cq;
    ep->ep_ops.create_sq = nvmf_emu_ctrl_create_sq;
    ep->ep_ops.delete_sq = nvmf_emu_ctrl_delete_sq;
    ep->ep_ops.activate_sq = nvmf_emu_ctrl_activate_sq;
    ep->ep_ops.deactivate_sq = nvmf_emu_ctrl_deactivate_sq;

    nvmx_info("NVMf ctrl created with type %d", ctrl->id.trtype);

    return 0;

free_state_lock:
    pthread_mutex_destroy(&ctrl->state_lock);
free_ctrl:
    nvmf_emu_free_ctrl(ctrl, nvme_emu_ep_trtype(ep->ep_type));
    return err;
}


void nvmf_emu_delete_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    // delete any namespace left in the list
    pthread_mutex_lock(&ctrl->ns_list_lock);
    nvmf_ctrl_clear_ns_list(ctrl);
    pthread_mutex_unlock(&ctrl->ns_list_lock);
    pthread_mutex_destroy(&ctrl->ns_list_lock);
    pthread_mutex_destroy(&ctrl->state_lock);
    nvmf_emu_free_ctrl(ctrl, ctrl->id.trtype);
}

static int nvmf_wait_ctrl_state(nvmf_emu_ctrl_t *ctrl, uint64_t cap, bool en)
{
    uint32_t ready_timeout_in_usec = 500000 * ((cap >> 24) & 0xff); //use regs.c func
    uint32_t csts;
    int err = -1;

    while (ready_timeout_in_usec > 0) {
        /* get Controller status */
        err = nvme_fabric_prop_get_cmd(ctrl, SNAP_NVME_REG_CSTS, NVMF_PROP_SIZE_4,
                                       &csts);
        if (err) {
            nvmx_error("get_csts() failed");
            break;
        }
        /* check ready bit */
        if ((le32toh(csts) & 0x1) == en) {
            err = 0;
            break;
        } else {
            ready_timeout_in_usec -= 500;
            usleep(500);
        }
    }

    if (ready_timeout_in_usec <= 0)
        err = -1;
    return err;
}

void nvmf_emu_fail_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    int ret;

    ret = sem_trywait(&ctrl->fail_sem);
    if (ret == -1 && errno == EAGAIN)
        /* Failover is already in progress */
        return;

    nvmx_info("ctrl %u: is failing on (%s:%s)", ctrl->cntlid, ctrl->ep->addr,
              ctrl->ep->port);
    nvme_emu_ep_fail_backend(ctrl->ep);
}

static int nvmf_keep_alive_send(nvmf_emu_ctrl_t *ctrl)
{
    nvmf_req_t *nvmf_req;
    nvme_cmd_t *cmd;
    int err;

    nvmx_debug("ctrl %u: sending Keep-alive", ctrl->cntlid);

    nvmf_req = ctrl->cmd_ops.get_request(ctrl, 0, false);
    if (!nvmf_req) {
        nvmx_error("ctrl %u: can't get new request for sending Keep-alive",
                   ctrl->cntlid);
        return -EAGAIN;
    }

    cmd = &nvmf_req->cmd;

    memset(cmd, 0, sizeof(*cmd));
    cmd->common.opc = NVME_ADMIN_KEEP_ALIVE;

    /* put async request only in case of a failure */
    err = ctrl->cmd_ops.send_cmd_async(ctrl, 0, nvmf_req, NULL, 0);
    if (err) {
        nvmx_error("ctrl %u: failed to send keep-alive cmd", ctrl->cntlid);
        ctrl->cmd_ops.put_request(ctrl, 0, nvmf_req);
    }

    return err;
}

static void nvmf_ctrl_handle_ns_changed_async_event(nvmf_emu_ctrl_t *ctrl)

{
    int rc;
    nvme_log_page_t log;
    uint32_t *changed_ns_list = log.changed_nslist_log.ns_list;

    /* sending log page re-arms target to send additional async events */
    rc = nvme_ctrl_get_log_page_cmd(ctrl, &log, NVME_NAMESPACE_GLOBAL,
                                    NVME_LOG_CHANGED_NS_LIST, 0);
    /*
     * There are two occasions where we might need to make a full
     * namespaces list rescan:
     * 1. If we failed to get the changed namespaces log page,
     *    possibly due to a non-supporting target.
     * 2. If num_ns_changes exceeds list size (1024) - this scenario is marked
     *    by target with 0xFFFFFFFF in first nsid entry.
     */
    if (rc || changed_ns_list[0] == NVME_NAMESPACE_GLOBAL) {
        nvmx_warn("ctrl %d: Cannot use Get log page of type %d, "
                  "running a full namespace list scan",
                  ctrl->cntlid, NVME_LOG_CHANGED_NS_LIST);
        nvmf_ctrl_update_namespaces(ctrl);
    } else {
        nvmf_ctrl_update_changed_namespaces(ctrl, changed_ns_list, 1024);
    }
}

static void nvmf_ctrl_handle_aen_notice(nvmf_emu_ctrl_t *ctrl, uint8_t ev_info)
{
    switch (ev_info) {
    case NVME_ASYNC_EVENT_NS_ATTR_CHANGED:
        nvmf_ctrl_handle_ns_changed_async_event(ctrl);
        break;
    default:
        nvmx_error("Unsupported async event info %d", ev_info);
    }
}

static void *nvmf_aen_thread(void *arg)
{
    nvmf_emu_ctrl_t *ctrl = arg;
    nvme_aen_completion_t event;
    int err;

    while (1) {
        sem_wait(&ctrl->aen_sem);

        /* Cache event before it is overridden by following events */
        memcpy(&event, &ctrl->aen_event, sizeof(nvme_aen_completion_t));

        /*
         * send new async event request, so target always keep
         * in-flight async event requests in its queue
         */
        err = nvme_ctrl_submit_async_event_cmd(ctrl);
        if (err)
            nvmx_warn("ctrl %d: Cannot send additional async event",
                      ctrl->cntlid);

        /* Handle Async event based on event parameters */
        switch (event.ev_type) {
        case NVME_ASYNC_EVENT_TYPE_NOTICE:
            nvmf_ctrl_handle_aen_notice(ctrl, event.ev_info);
            break;
        case NVME_ASYNC_EVENT_TYPE_ERROR:
        case NVME_ASYNC_EVENT_TYPE_SMART:
        case NVME_ASYNC_EVENT_TYPE_IO:
        case NVME_ASYNC_EVENT_TYPE_VENDOR:
            /* Fallthrough - we shouldn't get events that we don't support */
        default:
            nvmx_error("Unsupported async event type %d", event.ev_type);
        }
    }

    //Shouldn't ever get here
    return NULL;
}

static void *nvmf_keep_alive_thread(void *arg)
{
    nvmf_emu_ctrl_t *ctrl = arg;
    struct timespec ts;
    int err;

    while (1) {
        if (clock_gettime(CLOCK_REALTIME, &ts) == -1) {
            nvmx_error("ctrl %u: failed to get system time, exiting",
                       ctrl->cntlid);
            goto out;
        }
        ts.tv_sec += ctrl->ka_timeout_ms / 1000;
        ts.tv_nsec += (ctrl->ka_timeout_ms % 1000) * 1000000;

        err = sem_timedwait(&ctrl->ka_sem, &ts);
        if (err == -1 && errno == ETIMEDOUT) {
            nvmx_error("ctrl %u: Keep-Alive Timeout expired on (%s:%s) !",
                       ctrl->cntlid, ctrl->ep->addr, ctrl->ep->port);
            goto out;
        }
        /* the interval is timeout/2 */
        usleep(ctrl->ka_timeout_ms * 1000 / 2);
        err = nvmf_keep_alive_send(ctrl);
        if (err)
            goto out;
    }

out:
    nvmf_emu_fail_ctrl(ctrl);

    return NULL;
}

static void nvmf_stop_keep_alive(nvmf_emu_ctrl_t *ctrl)
{
    pthread_cancel(ctrl->kato_thread);
    pthread_join(ctrl->kato_thread, NULL);
}

static int nvmf_start_keep_alive(nvmf_emu_ctrl_t *ctrl)
{
    int err;

    err = pthread_create(&ctrl->kato_thread, NULL, nvmf_keep_alive_thread, ctrl);
    if (err) {
        nvmx_error("failed to start keep-alive thread for ctrl 0x%p", ctrl);
        return err;
    }

    /* send initial Keep-Alive packet */
    err = nvmf_keep_alive_send(ctrl);
    if (err)
        return err;

    return 0;
}

static int nvmf_set_num_io_queues(nvmf_emu_ctrl_t *ctrl)
{
    int remote_num_io_queues, num_io_queues = ctrl->nr_queues - 1;
    uint32_t dw11, res, sq_allocated, cq_allocated;
    int err;

    dw11 = ((num_io_queues - 1) << 16) | (num_io_queues - 1);
    err = nvme_ctrl_set_feature_cmd(ctrl, NVME_NUMBER_OF_QUEUES , dw11, &res);
    if (err)
        goto out;

    nvmx_debug("Set features - number of queues: sq %u cq %u",
               (res & 0xffff) + 1 , (res >> 16) + 1);
    /*
     * Lower 16-bits indicate number of submission queues allocated.
     * Upper 16-bits indicate number of completion queues allocated.
     */
    sq_allocated = (res & 0xffff) + 1;
    cq_allocated = (res >> 16) + 1;

    /*
     * For 1:1 queue mapping, set number of allocated queues to be minimum of
     * submission and completion queues.
     */
    remote_num_io_queues = nvmx_min(sq_allocated, cq_allocated);

    /*
     * We cannot allow target to not match initiator's num_io_queues,
     * since this value has already been promised to nvme host.
     */
    if (remote_num_io_queues < num_io_queues) {
        nvmx_error("Remote target cannot support %d IO queues (max %d)",
                   num_io_queues, remote_num_io_queues);
        err = -ENOMEM;
    }
out:
    return err;
}

static void nvmf_emu_disable_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    nvmf_stop_keep_alive(ctrl);
}

static int nvmf_emu_enable_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    int page_shift = 12;
    union nvme_cc_register *cc = &ctrl->cc;
    uint64_t cap;
    int err;

    /* get Controller Capabilities */
    err = nvme_fabric_prop_get_cmd(ctrl, SNAP_NVME_REG_CAP, NVMF_PROP_SIZE_8, &cap);
    if (err) {
        nvmx_error("ctrl %u: get_cap() failed err %d", ctrl->cntlid, err);
        return NVMX_ERR;
    }

    /* set page size to 4k by default */
    ctrl->page_sz = 1 << page_shift;

    cc->bits.en = 1;
    cc->bits.css = 0;
    cc->bits.shn = 0;
    cc->bits.ams = 0;
    cc->bits.iosqes = 6; /* SQ entry size == 64 == 2^6 */
    cc->bits.iocqes = 4; /* CQ entry size == 16 == 2^4 */
    /* Page size is 2 ^ (12 + mps). */
    cc->bits.mps = page_shift - 12;

    err = nvme_fabric_prop_set_cmd(ctrl, SNAP_NVME_REG_CC, NVMF_PROP_SIZE_4,
                                   cc->raw);
    if (err) {
        nvmx_error("ctrl %u: set cc failed err %d", ctrl->cntlid, err);
        return NVMX_ERR;
    }

    /* wait till ctrl is ready and enabled */
    err = nvmf_wait_ctrl_state(ctrl, cap, true);
    if (err)
        return err;

    err = nvmf_start_keep_alive(ctrl);
    if (err)
        // maybe disable ctrl to be a good citizen
        return err;
    return 0;
}

static void nvmf_copy_global_id_list(nvme_id_ns_descriptor_t id_descs[],
                                     char *list)
{
    nvme_id_ns_descriptor_t *id_desc = &id_descs[0];
    nvme_id_ns_descriptor_t *curr = (nvme_id_ns_descriptor_t *) list;
    int i = 0;

    /*
     * From the spec 5.15.5 - a Namespace Identifier Descriptor Length (NIDL)
     * value of 0h shall be interpreted as the end of the list
     */
    while (curr->nidl) {
        nvmx_debug("The target Global ID type: 0x%x", curr->nidt);

        switch (curr->nidt) {
            case NVME_NS_IDENTIFIER_EUI64:
            case NVME_NS_IDENTIFIER_NGUID:
            case NVME_NS_IDENTIFIER_UUID:
                memcpy(id_desc, curr, curr->nidl + 4);
                id_desc = &id_descs[++i];

                break;

                /*
                 * From the spec 5.15.5 - In case of the descriptor
                 * with unrecognized type just skip it and continue
                 * parsing the structure.
                 */
            default:
                break;
        }

        curr = (nvme_id_ns_descriptor_t *) ((char *) curr + curr->nidl + 4);
    }
}

int nvmf_identify_namespace(nvmf_emu_ctrl_t *ctrl, uint32_t nsid,
                            nvme_id_ns_t *id_ns,
                            nvme_id_ns_descriptor_t id_descs[])
{
    nvme_id_ns_t zero_id_ns = {0};
    int i, rc;

    rc = nvme_identify_cmd(ctrl, id_ns, sizeof(*id_ns), nsid,
                           NVME_IDENTIFY_CNS_NAMESPACE);
    if (rc)
        return rc;

    /*
     * From NVMe spec. (NVMe 1.3a, Figure 106):
     * "The Identify Namespace data structure is returned to the host for
     * the namespace specified in the Namespace Identifier (CDW1.NSID)
     * field if it is an active NSID. If the specified namespace is not
     * an active NSID, then the controller returns a zero filled
     * data structure."
     * From our POV, zero-filled data structure can be treated just like
     * a command failure.
     */
    if (!memcmp(id_ns, &zero_id_ns, sizeof(*id_ns)))
            return -EINVAL;

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i)
        memset(&id_descs[i], 0, sizeof(id_descs[i]));

    if (ctrl->ep->io_driver->ver >= NVME_VERSION(1, 3, 0) &&
        ctrl->cdata.ver.raw >= NVME_VERSION(1, 3, 0)) {
        static const int data_len = 4096;
        char *data;

        data = (char *) calloc(data_len, 1);
        if (!data)
            return -ENOMEM;

        rc = nvme_identify_cmd(ctrl, data, data_len, nsid,
                               NVME_IDENTIFY_CNS_NS_ID_DESCRIPTOR_LIST);
        if (rc) {
            nvmx_error("nsid %d: failed to get namespace ID descriptor list "
                       "from the target", nsid);
            free(data);
            return rc;
        }

        nvmf_copy_global_id_list(id_descs, data);

        free(data);
    } else {
       /*
        * In case the target doesn't support Namespace Identification
        * Descriptor list cmd we cannot get ALL global IDs relates to this
        * namespace and to check it vs. configuration, so we just
        * copy the first not zero ID from the id_ns structure
        * that can be used in case of dynamic namespace management.
        */
        uint8_t nguid[16] = {0};
        uint8_t eui64[8] = {0};
        nvme_id_ns_descriptor_t *id_desc = &id_descs[0];

        if (memcmp(id_ns->eui64, eui64, 8)) {
            id_desc->nidl = 8;
            id_desc->nidt = NVME_NS_IDENTIFIER_EUI64;

            memcpy(id_desc->nid.eui64.raw, &id_ns->eui64,
                   sizeof(id_ns->eui64));

            id_desc = &id_descs[1];
        }

        if (memcmp(id_ns->nguid, nguid, 16)) {
            id_desc->nidl = 16;
            id_desc->nidt = NVME_NS_IDENTIFIER_NGUID;

            memcpy(id_desc->nid.nguid.raw, &id_ns->nguid,
                   sizeof(id_ns->nguid));
        }
    }

    return 0;
}

int nvmf_identify_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    int err;

    /* get Controller Version */
    err = nvme_fabric_prop_get_cmd(ctrl, SNAP_NVME_REG_VS, NVMF_PROP_SIZE_4,
                                   &ctrl->vs);
    if (err) {
        nvmx_error("ctrl %u: get_vs() failed, err %d", ctrl->cntlid, err);
        goto out;
    }

    /* get Controller cap after READY state and save it */
    err = nvme_fabric_prop_get_cmd(ctrl, SNAP_NVME_REG_CAP, NVMF_PROP_SIZE_8,
                                   &ctrl->cap);
    if (err) {
        nvmx_error("ctrl %u: get_cap() failed, err %d", ctrl->cntlid, err);
        goto out;
    }

    nvmx_debug("CAP io_q depth %lu, default is %u", (ctrl->cap & 0xffff) + 1,
               ctrl->max_io_queue_size);
    ctrl->max_io_queue_size = nvmx_min(ctrl->cap & 0xffff,
                                       ctrl->max_io_queue_size - 1) + 1;

    err = nvme_identify_cmd(ctrl, &ctrl->cdata, sizeof(ctrl->cdata),
                            0, NVME_IDENTIFY_CNS_CTRL);
    if (err) {
        nvmx_error("ctrl %u: failed to identify", ctrl->cntlid);
        goto out;
    }

    ctrl->md_configurable = NVMF_CTRL_IS_MD_CONFIGURABLE(ctrl->cdata.mdts,
                                                         ctrl->ep->io_driver->mdts);
    return 0;

out:
    return err;
}

/* Must be called with ctrl->namespaces_lock held */
static int nvmf_ctrl_init_namespace(nvme_ns_t *ns, nvmf_emu_ctrl_t *ctrl,
                                    uint32_t nsid, const nvme_id_ns_t *id_ns,
                                    nvme_id_ns_descriptor_t id_descs[])
{
    nvme_id_ns_descriptor_t *ns_descs_data = ns->ns_descs_data;
    nvme_id_ns_t *nsdata = &ns->ns_data;
    int i, rc;

    memcpy(nsdata, id_ns, sizeof(*nsdata));

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i)
        memcpy(&ns_descs_data[i], &id_descs[i], sizeof(id_descs[i]));

    ns->ctrl = ctrl;
    ns->id = nsid;
    ns->block_order = nsdata->lbaf[NVME_ID_NS_FLBAS_INDEX(nsdata->flbas)].ds;
    ns->md_size = nsdata->lbaf[NVME_ID_NS_FLBAS_INDEX(nsdata->flbas)].ms;
    ns->size_b = (nsdata->nsze * (1 << ns->block_order));

    /* Check that there is no violation for NS with metadata */
    if (ns->md_size) {
        if (!ctrl->md_configurable) {
            nvmx_error("ns %d requires metadata which cannot be supported "
                       "with ctrl %d", nsid, ctrl->cntlid);
            rc = -EINVAL;
            goto clean_nsdata;
        }

        if (NVME_ID_NS_FLBAS_EXTENDED(nsdata->flbas)) {
            ns->pi_type = nsdata->dps & NVME_ID_NS_DPS_TYPE_MASK;
        } else {
            nvmx_error("ns %d doesn't support extended LBA for metadata",
                       ns->id);
            rc = -EINVAL;
            goto clean_nsdata;
        }
    }

    return 0;

clean_nsdata:
    memset(nsdata, 0, sizeof(*nsdata));

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i)
        memset(&ns_descs_data[i], 0, sizeof(ns_descs_data[i]));

    return rc;
}

/* Must be called with ctrl->namespaces_lock held */
static void nvmf_ctrl_attach_namespace(nvme_ns_t *ns,
                                       nvme_ns_list_t *ns_list)
{
    TAILQ_INSERT_TAIL(ns_list, ns, entry);
    nvme_emu_ep_attach_volume(ns->ctrl->ep, ns->id, ns->size_b,
                              ns->block_order, ns->md_size,
                              ns->ns_descs_data, NULL);
    nvmx_info("ctrl %d: Namespace nsid %d was added to NVMf namespace list",
              ns->ctrl->cntlid, ns->id);
}

/* Must be called with ctrl->namespaces_lock held */
static void nvmf_ctrl_detach_namespace(nvme_ns_t *ns,
                                       nvme_ns_list_t *ns_list)
{
    nvme_emu_ep_detach_volume(ns->ctrl->ep, ns->id);
    TAILQ_REMOVE(ns_list, ns, entry);
    nvmx_info("ctrl %d: Namespace nsid %d was removed from NVMf namespace list",
              ns->ctrl->cntlid, ns->id);
}

/* Must be called with ctrl->namespaces_lock held */
static int nvmf_ctrl_modify_namespace(nvme_ns_t *ns,
                                      const nvme_id_ns_t *id_ns,
                                      nvme_id_ns_descriptor_t id_descs[])
{
    int err;

    /*
     * In case the given id_ns data structure is different
     * than the cached value in nvme_ns_t struct, we need to
     * (re-)init the namespaces to its updated values (there
     * might be changes in disk size, block order, etc.)
     */
    if (memcmp(&ns->ns_data, id_ns, sizeof(ns->ns_data))) {
        err = nvmf_ctrl_init_namespace(ns, ns->ctrl, ns->id, id_ns, id_descs);
        if (err)
            goto err;
        nvmx_debug("ctrl %d: Namespace nsid %d was modified in NVMf namespace list",
                   ns->ctrl->cntlid, ns->id);
    }

    return 0;

err:
    return err;
}

/* Must be called with ctrl->namespaces_lock held */
static void nvmf_ctrl_delete_namespace(nvme_ns_t *ns)
{
    free(ns);
}

/* Must be called with ctrl->namespaces_lock held */
static nvme_ns_t *nvmf_ctrl_create_namespace(nvmf_emu_ctrl_t *ctrl,
                                             uint32_t nsid,
                                             const nvme_id_ns_t *id_ns,
                                             nvme_id_ns_descriptor_t id_descs[])
{
    int rc;
    nvme_ns_t *ns;

    ns = calloc(1, sizeof(nvme_ns_t));
    if (!ns)
        goto err;

    rc = nvmf_ctrl_init_namespace(ns, ctrl, nsid, id_ns, id_descs);
    if (rc)
        goto err_free_ns;

    return ns;

err_free_ns:
    free(ns);
err:
    return NULL;
}

/* Must be called with ctrl->namespaces_lock held */
static nvme_ns_t *nvmf_ctrl_find_namespace(nvmf_emu_ctrl_t *ctrl, uint32_t nsid)
{
    nvme_ns_t *ns;

    TAILQ_FOREACH(ns, &ctrl->ns_list, entry) {
        if (ns->id == nsid)
            return ns;
    }
    return NULL;
}

static nvme_ns_t *nvmf_ctrl_find_namespace_lock(nvmf_emu_ctrl_t *ctrl, uint32_t nsid)
{
    nvme_ns_t *ns;
    pthread_mutex_lock(&ctrl->ns_list_lock);
    ns = nvmf_ctrl_find_namespace(ctrl, nsid);
    pthread_mutex_unlock(&ctrl->ns_list_lock);
    return ns;
}

static int nvmf_ctrl_update_namespace(nvmf_emu_ctrl_t *ctrl, uint32_t nsid,
                                      nvme_ns_list_t *new_ns_list)
{
    nvme_ns_t *ns;
    nvme_id_ns_t id_ns;
    nvme_id_ns_descriptor_t id_descs[NVME_NS_ID_TYPES_NUM];
    int rc;

    memset(&id_ns, 0, sizeof(id_ns));

    rc = nvmf_identify_namespace(ctrl, nsid, &id_ns, id_descs);
    if (rc) {
        nvmx_warn("Failed to identify namespace %u", nsid);
        pthread_mutex_lock(&ctrl->ns_list_lock);
        if ((ns = nvmf_ctrl_find_namespace(ctrl, nsid)) != NULL) {
            nvmf_ctrl_detach_namespace(ns, &ctrl->ns_list);
            nvmf_ctrl_delete_namespace(ns);
        }
        pthread_mutex_unlock(&ctrl->ns_list_lock);
        goto out;
    }

    pthread_mutex_lock(&ctrl->ns_list_lock);
    ns = nvmf_ctrl_find_namespace(ctrl, nsid);
    if (!ns) {
        /* A new namespace has been discovered */
        ns = nvmf_ctrl_create_namespace(ctrl, nsid, &id_ns, id_descs);
        if (!ns) {
            nvmx_warn("Failed to create namespace %d", nsid);
            rc = -ENOMEM;
            goto out_unlock;
        }
    } else {
        /* Namespace is already configured, might need to modify */
        nvmf_ctrl_detach_namespace(ns, &ctrl->ns_list);
        rc = nvmf_ctrl_modify_namespace(ns, &id_ns, id_descs);
        if (rc) {
            nvmx_warn("Failed to modify namespace %d", nsid);
            nvmf_ctrl_delete_namespace(ns);
            goto out_unlock;
        }
    }
    nvmf_ctrl_attach_namespace(ns, new_ns_list);

out_unlock:
    pthread_mutex_unlock(&ctrl->ns_list_lock);
out:
    return rc;
}

static void nvmf_ctrl_update_namespaces(nvmf_emu_ctrl_t *ctrl)
{
    uint32_t i, j;
    uint32_t nsid = 0;
    uint32_t max_num_ns_per_iter;
    nvme_ns_list_t new_ns_list;
    nvme_nsid_list_t nsid_list;
    int rc = 0;

    max_num_ns_per_iter = sizeof(nsid_list.ns_list) / sizeof(nsid_list.ns_list[0]);
    TAILQ_INIT(&new_ns_list);
    if (ctrl->cdata.ver.raw >= NVME_VERSION(1, 1, 0)) {
        while (true) {
           /*
            * Iterate through the pages and fetch each chunk of 1024 namespaces until
            * there are no more active namespaces
            */
            memset(&nsid_list, 0, sizeof(nsid_list));
            rc = nvme_identify_cmd(ctrl, &nsid_list.ns_list,
                                   sizeof(nsid_list.ns_list),
                                   nsid,
                                   NVME_IDENTIFY_CNS_ACTIVE_NS_LIST);
            if (rc) {
                nvmx_warn("Failed to Identify Namespace list");
                break;
            }

            for (j = 0; j < max_num_ns_per_iter; j++) {
                nsid = le32toh(nsid_list.ns_list[j]);
                if (nsid == 0)
                    break;
                /*
                 * If namespace is failed to be updated, move forward to next
                 * nsid. All mandatory namespaces are being verified later.
                 */
                nvmf_ctrl_update_namespace(ctrl, nsid, &new_ns_list);
            }

            /* if we encounter a zero ns id, it means no updates anymore */
            if (nsid == 0)
                break;
        }
    } else {
        /*
         * Controller doesn't support active ns list CNS 0x02 so identify all
         * nsid range up to cdata.nn (max nsid).
         */
        nvmx_warn("Target doesn't support ns_list, scan all nsid range");
        for (i = 0; i < ctrl->cdata.nn; i++) {
            /*
             * If namespace is failed to be updated, move forward to next
             * nsid. All mandatory namespaces are being verified later.
             */
            nvmf_ctrl_update_namespace(ctrl, nsid, &new_ns_list);
        }
    }

    /* Remove dead namespaces, and append the new list instead */
    pthread_mutex_lock(&ctrl->ns_list_lock);
    nvmf_ctrl_clear_ns_list(ctrl);
    TAILQ_CONCAT(&ctrl->ns_list, &new_ns_list, entry);
    pthread_mutex_unlock(&ctrl->ns_list_lock);
}

static void nvmf_ctrl_update_changed_namespaces(nvmf_emu_ctrl_t *ctrl,
                                                uint32_t *ns_list,
                                                size_t ns_list_size)
{
    int i;
    uint32_t nsid;

    for (i = 0; i < ns_list_size; i++) {
        nsid = ns_list[i];
        if (!nsid)
            break;

        nvmf_ctrl_update_namespace(ctrl, nsid, &ctrl->ns_list);
    }
}

/* Must be called with ctrl->namespaces_lock held */
static void nvmf_ctrl_clear_ns_list(nvmf_emu_ctrl_t *ctrl)
{
    nvme_ns_t *ns;

    while ((ns = TAILQ_FIRST(&ctrl->ns_list)) != NULL) {
        nvmf_ctrl_detach_namespace(ns, &ctrl->ns_list);
        nvmf_ctrl_delete_namespace(ns);
    }
    nvmx_assertv_always(TAILQ_EMPTY(&ctrl->ns_list), "ns list not cleared");
}

static int nvmf_ctrl_verify_mdts_configuration(nvmf_emu_ctrl_t *ctrl)
{
    int host_mdts = ctrl->ep->io_driver->mdts;
    int target_mdts = ctrl->cdata.mdts;
    int rc = 0;

    /* Remote target is not limited in mdts value (mdts == 0) */
    if (!target_mdts)
        return 0;

    if (host_mdts == 0) {
        nvmx_error("Remote target mdts is limited to %d while "
                   "configured mdts is unlimited (valued to %d)",
                   target_mdts, host_mdts);
        rc = -EINVAL;
    } else if (target_mdts < host_mdts) {
        nvmx_error("Remote target mdts (%d) cannot be smaller than "
                   "configured mdts (%d)", target_mdts, host_mdts);
        rc = -EINVAL;
    }

    return rc;
}

static int nvmf_ctrl_verify_oncs_configuration(nvmf_emu_ctrl_t *ctrl,
                                               snap_conf_t *sconfig)

{
    uint16_t oncs;

    /* if no config file, we don't need to verify anything */
    if (!sconfig)
        return NVMX_OK;

    /*
     * ONCS is 0 in case there is no oncs field in the config,
     * so it doesn't matter if some bit is set on the target.
     */
    oncs = snap_config_get_oncs(sconfig);

    if (oncs && !(oncs & ctrl->cdata.oncs.raw)) {
        nvmx_error("according oncs configuration some requested "
                   "features are not supported by the target: "
                   "target ONCS %u, from conf file %u",
                   ctrl->cdata.oncs, oncs);
        return NVMX_ERR_NOTSUPPORTED;
    }

    return NVMX_OK;
}

static int nvmf_ctrl_verify_ctrl_configuration(nvmf_emu_ctrl_t *ctrl,
                                               snap_conf_t *sconfig)
{
    int err;

    if (ctrl->ep->io_driver->ver >= NVME_VERSION(1, 1, 0) &&
        ctrl->cdata.ver.raw < NVME_VERSION(1, 1, 0)) {
        nvmx_error("ctrl %u: target ver %d is less than "
                   "the ver in the config file %d that is not supported "
                   "for dynamic namespace management",
                   ctrl->cntlid, ctrl->cdata.ver, ctrl->ep->io_driver->ver);
        return NVMX_ERR_NOTSUPPORTED;
    }

    err = nvmf_ctrl_verify_oncs_configuration(ctrl, sconfig);
    if (err)
        return err;

    err = nvmf_ctrl_verify_mdts_configuration(ctrl);
    if (err)
        return err;

    return 0;
}

static void nvmf_ctrl_deconfigure_async_events(nvmf_emu_ctrl_t *ctrl)
{
    if (!(le32toh(ctrl->cdata.oaes.raw) & NVMF_SUPPORTED_ASYNC_EVENTS))
        return;

    pthread_cancel(ctrl->aen_thread);
    pthread_join(ctrl->aen_thread, NULL);
    sem_destroy(&ctrl->aen_sem);
}

static int nvmf_ctrl_configure_async_events(nvmf_emu_ctrl_t *ctrl)
{
    uint32_t res;
    int err;
    uint32_t supported_aes;

    supported_aes = le32toh(ctrl->cdata.oaes.raw) & NVMF_SUPPORTED_ASYNC_EVENTS;
    if (!supported_aes)
        return 0;

    err = nvme_ctrl_set_feature_cmd(ctrl, NVME_ASYNCHRONOUS_EVENT_CONF,
                                    supported_aes, &res);
    if (err) {
        /*
         * Old linux kernel targets (older than v4.18) support some
         * async events, but does not support NVME_ASYNCHRONOUS_EVENT_CONF
         * set features command. We don't want to fail connection to
         * this kind of targets, only ignore events we don't really support.
         */
        nvmx_warn("Cannot configure target's supported async events (0x%x)",
                   supported_aes);
    }

    err = sem_init(&ctrl->aen_sem, 0, 0);
    if (err)
        goto err;

    err = pthread_create(&ctrl->aen_thread, NULL, nvmf_aen_thread, ctrl);
    if (err) {
        nvmx_warn("Failed to create async events handling thread");
        goto sem_destroy;
    }

    err = nvme_ctrl_submit_async_event_cmd(ctrl);
    if (err) {
        nvmx_warn("Failed to send async event request");
        goto pthread_destroy;
    }

    return 0;

pthread_destroy:
    pthread_cancel(ctrl->aen_thread);
    pthread_join(ctrl->aen_thread, NULL);
sem_destroy:
    sem_destroy(&ctrl->aen_sem);
err:
    return err;
}
