#include <poll.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <linux/if_ether.h>

#include "config.h"
#include "nvmf_rdma.h"
#include "nvme_emu_io_driver.h"

static void nvmf_rdma_destroy_transport_resources(nvmf_rdma_qpair_t *queue);
static int nvmf_rdma_post_recv(nvmf_rdma_qpair_t *queue,
                               struct ibv_recv_wr *wr);

static void nvmf_rdma_put_req(nvmf_rdma_qpair_t *queue,
                              struct nvmf_rdma_req *rdma_req)
{
    pthread_spin_lock(&queue->reqs_lock);
    TAILQ_REMOVE(&queue->outstanding_reqs, rdma_req, entry);
    TAILQ_INSERT_HEAD(&queue->free_reqs, rdma_req, entry);
    pthread_spin_unlock(&queue->reqs_lock);
}

static struct nvmf_rdma_req *nvmf_rdma_get_req(nvmf_rdma_qpair_t *queue,
                                               bool sync)
{
    struct nvmf_rdma_req *rdma_req;

    pthread_spin_lock(&queue->reqs_lock);
    rdma_req = TAILQ_FIRST(&queue->free_reqs);
    if (rdma_req) {
        TAILQ_REMOVE(&queue->free_reqs, rdma_req, entry);
        TAILQ_INSERT_TAIL(&queue->outstanding_reqs, rdma_req, entry);
        rdma_req->nvmf_req.is_sync = sync;
        rdma_req->nvmf_req.payload_size = 0;
        rdma_req->nvmf_req.status = NULL;
    }
    pthread_spin_unlock(&queue->reqs_lock);

    return rdma_req;

}

static void nvmf_rdma_put_request(nvmf_emu_ctrl_t *ctrl, uint16_t qid,
                                  nvmf_req_t *nvmf_req)
{
    nvmf_rdma_ctrl_t *rctrl = to_rdma_ctrl(ctrl);
    nvmf_rdma_qpair_t *queue = &rctrl->queues[qid];
    struct nvmf_rdma_req *rdma_req = to_rdma_req(nvmf_req);

    nvmf_rdma_put_req(queue, rdma_req);
}

static nvmf_req_t* nvmf_rdma_get_request(nvmf_emu_ctrl_t *ctrl, uint16_t qid,
                                         bool sync)
{
    nvmf_rdma_ctrl_t *rctrl = to_rdma_ctrl(ctrl);
    nvmf_rdma_qpair_t *queue = &rctrl->queues[qid];
    struct nvmf_rdma_req *rdma_req;

    rdma_req = nvmf_rdma_get_req(queue, sync);
    if (!rdma_req)
        return NULL;
    else
        return &rdma_req->nvmf_req;
}

static int nvmf_rdma_process_rsp(struct nvmf_rdma_rsp *rsp,
                                 struct nvmf_rdma_req *req)
{
    nvme_cmd_t *cmd = &req->nvmf_req.cmd;
    nvme_cqe_t *cqe = &rsp->cqe;
    uint8_t opcode = cmd->common.opc;
    nvmf_emu_ctrl_t *ctrl = &req->queue->ctrl->ctrl;

    switch (opcode) {
    case NVME_ADMIN_KEEP_ALIVE:
        sem_post(&ctrl->ka_sem);
        break;
    case NVME_ADMIN_ASYNC_EV_REQ:
        memcpy(&ctrl->aen_event, &cqe->result, sizeof(nvme_aen_completion_t));
        sem_post(&ctrl->aen_sem);
        break;
    default:
        nvmx_debug("got opcode %u", opcode);
        if (req->nvmf_req.status)
            memcpy(&req->nvmf_req.status->cqe, cqe, sizeof(nvme_cqe_t));
        break;
    }

    return 0;
}

static int nvmf_rdma_recv(nvmf_rdma_qpair_t *queue, uint64_t rsp_idx)
{
    struct nvmf_rdma_rsp *rdma_rsp = &queue->rdma_rsps[rsp_idx];
    struct nvmf_rdma_req *rdma_req = &queue->rdma_reqs[rdma_rsp->cqe.cid];
    int ret;

    ret = nvmf_rdma_process_rsp(rdma_rsp, rdma_req);

    if (rdma_req->nvmf_req.is_sync)
        sem_post(&rdma_req->nvmf_req.sync);
    else
        nvmf_rdma_put_req(queue, rdma_req);

    if (!ret)
         ret = nvmf_rdma_post_recv(queue, &rdma_rsp->recv_wr);

    return ret;
}

static int nvmf_rdma_handle_completion(struct ibv_wc *wc,
                                       nvmf_rdma_qpair_t *queue)
{
    int ret;

    if (wc->status == IBV_WC_SUCCESS) {
        switch (wc->opcode) {
            case IBV_WC_RECV:
                if (unlikely(wc->byte_len < sizeof(nvme_cqe_t))) {
                    nvmx_error("recv length %u less than expected rsp size",
                               wc->byte_len);
                    return -1;
                }
                ret = nvmf_rdma_recv(queue, wc->wr_id);
                if (unlikely(ret)) {
                    nvmx_error("nvmf_rdma_recv processing failure");
                    return -1;
                }
                break;
            case IBV_WC_SEND:
                /* nothing to do in send completion */
                break;
            default:
                nvmx_error("Received an unexpected completion with opcode: %d",
                           wc->opcode);
                return -1;
        }
    } else if (wc->status == IBV_WC_WR_FLUSH_ERR) {
        nvmx_debug("received IBV_WC_WR_FLUSH_ERR comp status");
    } else {
        nvmx_info("received %d comp status", wc->status);
    }

    return 0;
}

static int nvmf_rdma_poll_cq(nvmf_rdma_qpair_t *queue)
{
    struct ibv_cq *cq = queue->cq;
    struct ibv_wc wc[NVMF_RDMA_POOL_BATCH];
    int n, i, completed = 0;
    int ret;

    while ((n = ibv_poll_cq(cq, NVMF_RDMA_POOL_BATCH, wc)) > 0) {
        if (n < 0) {
            nvmx_error("queue %d: ibv_poll_cq() error", queue->qid);
            return -1;
        }

        for (i = 0; i < n; i++) {
            ret = nvmf_rdma_handle_completion(&wc[i], queue);
            if (ret)
                return -1;
        }

        completed += n;
        if (completed >= NVMF_RDMA_MAX_COMPLETIONS)
            goto out;
    }

    if (n > 0) {
        nvmx_error("still have %d elements to process for queue %d", n,
                   queue->qid);
        return -1;
    }

out:
    return 0;
}

/*
 * When timeout < 0 the function will poll without timeout. This is used
 * for cases in which the timeout is triggred by the rdma_cm (e.g. when
 * performing rdma_resolve_addr)
 */
static struct rdma_cm_event *nvmf_rdma_get_cm_event(
                                struct rdma_event_channel *channel,
                                enum rdma_cm_event_type expected_evt,
                                int timeout)
{
    struct rdma_cm_event *event;
    int rc;
    struct pollfd pfd = {0};


    pfd.fd = channel->fd;
    pfd.events = POLLIN;
    pfd.revents = 0;
    rc = poll(&pfd, 1, timeout);
    if (rc < 0) {
        nvmx_error("expected event %s but failed polling on event channel "
                   "with error: %s", rdma_event_str(expected_evt),
                   strerror(errno));
        return NULL;
    }
    if (rc == 0) {
        nvmx_error("timeout while polling on event channel, expected event"
                   " %s", rdma_event_str(expected_evt));
        return NULL;
    }

    rc = rdma_get_cm_event(channel, &event);
    if (rc) {
        nvmx_error("Failed to get event %s from CM event channel. Error"
                   " %d (%s)", rdma_event_str(expected_evt), errno,
                   strerror(errno));
        return NULL;
    }

    if (event->event != expected_evt) {
        nvmx_error("Expected %s but received %s (%d) from CM event"
                    " channel (status = %d)", rdma_event_str(expected_evt),
                    rdma_event_str(event->event), event->event, event->status);
        rdma_ack_cm_event(event);
        return NULL;
    }

    return event;
}

static int nvmf_rdma_set_channel_fd_nonblock(struct rdma_event_channel *channel)
{
    int flags;
    int ret;

    flags = fcntl(channel->fd, F_GETFL);
    ret = fcntl(channel->fd, F_SETFL, flags | O_NONBLOCK);
    if (ret < 0) {
        nvmx_error("failed to set async fd to nonblock");
        return -1;
    }

    return 0;
}

static int __nvmf_rdma_cq_handler(nvmf_rdma_qpair_t *queue)
{
    struct ibv_cq *ev_cq;
    void *ev_ctx;
    int ret;

    ret = ibv_get_cq_event(queue->comp_channel, &ev_cq, &ev_ctx);
    if (ret) {
        nvmx_error("failed to get cq event for qid %d", queue->qid);
        goto out;
    }

    if (ev_cq != queue->cq) {
        nvmx_error("Got event on unknown CQ");
        ret = EAGAIN;
        goto out;
    }

    ret = ibv_req_notify_cq(queue->cq, 0);
    if (ret) {
        nvmx_error("failed to notify cq 0x%p", queue->cq);
        goto out;
    }

    ibv_ack_cq_events(queue->cq, 1);

    ret = nvmf_rdma_poll_cq(queue);
    if (ret)
        goto out;

    return 0;

out:
    return ret;
}

static void *cq_thread(void *arg)
{
    nvmf_rdma_qpair_t *queue = arg;
    int ret;

    while (1) {
        ret = __nvmf_rdma_cq_handler(queue);
        if (ret)
            pthread_exit(NULL);
    }

    return NULL;
}

static void nvmf_rdma_destroy_cq(nvmf_rdma_qpair_t *queue)
{
    pthread_cancel(queue->cq_thread);
    pthread_join(queue->cq_thread, NULL);
    ibv_destroy_cq(queue->cq);
    ibv_destroy_comp_channel(queue->comp_channel);
}

static int nvmf_rdma_create_cq(nvmf_rdma_qpair_t *queue)
{
    int err;
    struct ibv_context *ibv_context = queue->cm_id->verbs;

    queue->comp_channel = ibv_create_comp_channel(ibv_context);
    if (!queue->comp_channel) {
        nvmx_error("ibv_create_comp_channel failed 0x%p - %m", queue);
        err = -errno;
        goto out;
    }

    queue->cq = ibv_create_cq(queue->cm_id->verbs, 2 * queue->queue_depth + 1,
                              queue, queue->comp_channel,
                              queue->qid % ibv_context->num_comp_vectors);
    if (!queue->cq) {
        nvmx_error("ibv_create_cq failed 0x%p qid %d - %m",
                   queue, queue->qid);
        err = -errno;
        goto free_channel;
    }

    err = ibv_req_notify_cq(queue->cq, 0);
    if (err) {
        nvmx_error("ibv_req_notify_cq failed 0x%p qid %d", queue, queue->qid);
        goto free_cq;
    }

    err = pthread_create(&queue->cq_thread, NULL, cq_thread, queue);
    if (err)
        goto free_cq;

    return 0;

free_cq:
    ibv_destroy_cq(queue->cq);
free_channel:
    ibv_destroy_comp_channel(queue->comp_channel);
out:
    return err;
}
static int nvmf_rdma_post_recv(nvmf_rdma_qpair_t *queue,
                               struct ibv_recv_wr *wr)
{
    struct ibv_recv_wr *bad_wr = NULL;
    int rc = 0;

    rc = ibv_post_recv(queue->qp, wr, &bad_wr);
    if (unlikely(rc))
        nvmx_error("Failure posting rdma recv, rc = 0x%x on queue 0x%p",
                   rc, queue);

    return rc;
}

static void nvmf_rdma_free_rsps(nvmf_rdma_qpair_t *queue)
{
    ibv_dereg_mr(queue->rdma_rsps_mr);
    free(queue->rdma_rsps);
}

static int nvmf_rdma_alloc_rsps(nvmf_rdma_qpair_t *queue)
{
    int ret;
    int i;

    queue->rdma_rsps = calloc(queue->queue_depth, sizeof(*queue->rdma_rsps));
    if (!queue->rdma_rsps) {
        nvmx_error("Failed to allocate RDMA rsps for queue 0x%p", queue);
        ret = -ENOMEM;
        goto fail;
    }

    queue->rdma_rsps_mr = ibv_reg_mr(queue->ctrl->pd, &queue->rdma_rsps,
                                      queue->queue_depth * sizeof(struct nvmf_rdma_rsp),
                                      IBV_ACCESS_LOCAL_WRITE);
    if (!queue->rdma_rsps_mr) {
        nvmx_error("Failed to reg cqe_mr for queue 0x%p", queue);
        ret = -EINVAL;
        goto out_free_queue_rsps;
    }

    for (i = 0; i < queue->queue_depth; i++) {
        struct nvmf_rdma_rsp *rdma_rsp = &queue->rdma_rsps[i];
        struct ibv_sge *recv_sgl = &rdma_rsp->recv_sgl;
        struct ibv_recv_wr *recv_wr = &rdma_rsp->recv_wr;

        rdma_rsp->cqe_mr = queue->rdma_rsps_mr;

        recv_sgl->addr = (uint64_t)&rdma_rsp->cqe;
        recv_sgl->length = sizeof(rdma_rsp->cqe);
        recv_sgl->lkey = rdma_rsp->cqe_mr->lkey;

        recv_wr->wr_id = i;
        recv_wr->next = NULL;
        recv_wr->sg_list = recv_sgl;
        recv_wr->num_sge = 1;

        if (nvmf_rdma_post_recv(queue, recv_wr)) {
            ret = -EINVAL;
            goto out_free_rsps;
        }
    }

    return 0;

out_free_rsps:
    ibv_dereg_mr(queue->rdma_rsps_mr);
out_free_queue_rsps:
    free(queue->rdma_rsps);
fail:
    return ret;
}

static void nvmf_rdma_free_reqs(nvmf_rdma_qpair_t *queue)
{
    int i;

    for (i = 0; i < queue->queue_depth; i++) {
        struct nvmf_rdma_req *rdma_req = &queue->rdma_reqs[i];

        TAILQ_REMOVE(&queue->free_reqs, rdma_req, entry);
        sem_destroy(&rdma_req->nvmf_req.sync);
    }

    pthread_spin_destroy(&queue->outstanding_reqs_lock);
    pthread_spin_destroy(&queue->free_reqs_lock);
    pthread_spin_destroy(&queue->reqs_lock);
    ibv_dereg_mr(queue->total_payload_mr);
    free(queue->total_payload);
    ibv_dereg_mr(queue->rdma_reqs_mr);
    free(queue->rdma_reqs);
}

static int nvmf_rdma_alloc_reqs(nvmf_rdma_qpair_t *queue)
{
    int i, j;

    queue->rdma_reqs = calloc(queue->queue_depth, sizeof(struct nvmf_rdma_req));
    if (!queue->rdma_reqs) {
        nvmx_error("Failed to allocate rdma_reqs for queue 0x%p", queue);
        goto fail;
    }

    queue->rdma_reqs_mr = ibv_reg_mr(queue->ctrl->pd,
                                queue->rdma_reqs,
                                queue->queue_depth * sizeof(struct nvmf_rdma_req),
                                IBV_ACCESS_LOCAL_WRITE);
    if (!queue->rdma_reqs_mr) {
        nvmx_error("Failed to reg cmd_mr for for queue 0x%p", queue);
        goto out_free_queue_reqs;
    }

    queue->total_payload = calloc(queue->queue_depth, NVMF_RDMA_PAYLOAD_BUFFER_SIZE);
    if (!queue->total_payload) {
        nvmx_error("Failed to allocate buffer for cmds queue 0x%p", queue);
        goto out_dereg_reqs;
    }

    queue->total_payload_mr = ibv_reg_mr(queue->ctrl->pd,
                                    queue->total_payload,
                                    queue->queue_depth * NVMF_RDMA_PAYLOAD_BUFFER_SIZE,
                                    IBV_ACCESS_LOCAL_WRITE |
                                    IBV_ACCESS_REMOTE_READ |
                                    IBV_ACCESS_REMOTE_WRITE);
    if (!queue->total_payload_mr) {
        nvmx_error("Failed to reg payload_mr for cmd queue 0x%p", queue);
        goto out_free_total_payload;
    }

    pthread_spin_init(&queue->reqs_lock, 0);
    TAILQ_INIT(&queue->free_reqs);
    pthread_spin_init(&queue->free_reqs_lock, 0);
    TAILQ_INIT(&queue->outstanding_reqs);
    pthread_spin_init(&queue->outstanding_reqs_lock, 0);

    for (i = 0; i < queue->queue_depth; i++) {
        struct nvmf_rdma_req    *rdma_req;
        nvme_cmd_t              *cmd;

        rdma_req = &queue->rdma_reqs[i];
        cmd = &rdma_req->nvmf_req.cmd;

        if (sem_init(&rdma_req->nvmf_req.sync, 0, 0)) {
            nvmx_error("Failed to init semaphore: %m");
            goto out_free_reqs;
        }

        rdma_req->cmd_mr = queue->rdma_reqs_mr;
        rdma_req->nvmf_req.payload = queue->total_payload + i * NVMF_RDMA_PAYLOAD_BUFFER_SIZE;
        rdma_req->payload_mr = queue->total_payload_mr;

        rdma_req->nvmf_req.id = i;
        rdma_req->queue = queue;

        rdma_req->send_sgl.addr = (uint64_t)cmd;
        rdma_req->send_sgl.length = sizeof(*cmd);
        rdma_req->send_sgl.lkey = rdma_req->cmd_mr->lkey;

        rdma_req->send_wr.wr_id = (uint64_t)rdma_req;
        rdma_req->send_wr.next = NULL;
        rdma_req->send_wr.opcode = IBV_WR_SEND;
        rdma_req->send_wr.send_flags = IBV_SEND_SIGNALED;
        rdma_req->send_wr.sg_list = &rdma_req->send_sgl;
        rdma_req->send_wr.num_sge = 1;
        rdma_req->send_wr.imm_data = 0;

        TAILQ_INSERT_TAIL(&queue->free_reqs, rdma_req, entry);
    }

    return 0;

out_free_reqs:
    for (j = 0; j < i; j++) {
        struct nvmf_rdma_req *rdma_req = &queue->rdma_reqs[j];

        TAILQ_REMOVE(&queue->free_reqs, rdma_req, entry);
        sem_destroy(&rdma_req->nvmf_req.sync);
    }
    pthread_spin_destroy(&queue->outstanding_reqs_lock);
    pthread_spin_destroy(&queue->free_reqs_lock);
    ibv_dereg_mr(queue->total_payload_mr);
out_free_total_payload:
    free(queue->total_payload);
out_dereg_reqs:
    ibv_dereg_mr(queue->rdma_reqs_mr);
out_free_queue_reqs:
    free(queue->rdma_reqs);
fail:
    return ENOMEM;
}

static void nvmf_rdma_destroy_qp(nvmf_rdma_qpair_t *queue)
{
    rdma_destroy_qp(queue->cm_id);
}

static int nvmf_rdma_create_qp(nvmf_rdma_qpair_t *queue)
{
    int	ret;
    struct ibv_qp_init_attr attr;

    memset(&attr, 0, sizeof(struct ibv_qp_init_attr));
    attr.qp_type = IBV_QPT_RC;
    attr.qp_context = queue;
    attr.send_cq = queue->cq;
    attr.recv_cq = queue->cq;
    attr.cap.max_send_wr = queue->queue_depth; /* SEND operations */
    attr.cap.max_recv_wr = queue->queue_depth; /* RECV operations */
    attr.cap.max_send_sge = 1; /* TODO: add support for inline data */
    attr.cap.max_recv_sge = 1;

    ret = rdma_create_qp(queue->cm_id, queue->ctrl->pd, &attr);
    if (ret) {
        nvmx_error("rdma_create_qp failed on %d", queue->qid);
        return ret;
    }
    queue->qp = queue->cm_id->qp;
    nvmx_info("QPN 0x%x created for qid %d", queue->qp->qp_num, queue->qid);
    return 0;
}

static void nvmf_rdma_destroy_transport_resources(nvmf_rdma_qpair_t *queue)
{
    nvmx_debug("queue %d: destroying transport resources", queue->qid);

    nvmf_rdma_destroy_qp(queue);
    nvmf_rdma_destroy_cq(queue);
    if (!queue->qid)
        ibv_dealloc_pd(queue->ctrl->pd);

}

static int nvmf_rdma_create_transport_resources(nvmf_rdma_qpair_t *queue)
{
    nvmf_rdma_ctrl_t *ctrl = queue->ctrl;
    int err;

    /* Create PD per ctrl, do it during admin queue creation */
    if (!queue->qid) {
        ctrl->pd = ibv_alloc_pd(queue->cm_id->verbs);
        if (!ctrl->pd) {
            nvmx_error("failed to allocate pd for ctrl %p", ctrl);
            err = ENOMEM;
            goto out;
        }
    }

    err = nvmf_rdma_create_cq(queue);
    if (err)
        goto dealloc_pd;

    err = nvmf_rdma_create_qp(queue);
    if (err)
        goto destroy_cq;

    return 0;

destroy_cq:
    nvmf_rdma_destroy_cq(queue);
dealloc_pd:
    if (!queue->qid)
        ibv_dealloc_pd(ctrl->pd);
out:
    return err;
}

/*
 * Must be called with state_lock held
 */
static int nvmf_rdma_resolve_route(nvmf_rdma_qpair_t *queue)
{
    int err;
    struct rdma_cm_event *event;

    err = nvmf_rdma_create_transport_resources(queue);
    if (err) {
        nvmx_error("queue %d: failed to create transport resources",
                   queue->qid);
        return err;
    }

    err = rdma_resolve_route(queue->cm_id, NVME_RDMA_TIMEOUT_IN_MS);
    if (err) {
        nvmx_error("queue %d: rdma_resolve_route failed", queue->qid);
        goto destroy_transport;
    }

    event = nvmf_rdma_get_cm_event(queue->cm_channel,
                                   RDMA_CM_EVENT_ROUTE_RESOLVED, -1);
    if (event == NULL) {
        nvmx_error("queue %d: failed to receive expected event", queue->qid);
        err = -1;
        goto destroy_transport;
    }
    rdma_ack_cm_event(event);

    return 0;

destroy_transport:
    nvmf_rdma_destroy_transport_resources(queue);
    return err;
}

/*
 * Must be called with state_lock held
 */
static int _nvmf_rdma_connect(nvmf_rdma_qpair_t *queue)
{
    struct rdma_conn_param param = {0};
    struct nvmf_rdma_request_private_data request_data = {0};
    struct ibv_device_attr attr;
    int ret;

    ret = ibv_query_device(queue->cm_id->verbs, &attr);
    if (ret) {
        nvmx_error("Failed to query RDMA device attributes.");
        goto out;
    }

    /* Set PI capability during admin queue resolution */
    if (!queue->qid)
        //queue->ctrl->ctrl.pi_capable = !!(attr.device_cap_flags & IBV_DEVICE_SIGNATURE_HANDOVER);
        queue->ctrl->ctrl.pi_capable = !!(attr.device_cap_flags & (1 << 30));

    param.responder_resources = nvmx_min(queue->queue_depth, attr.max_qp_rd_atom);
    param.private_data = &request_data;
    param.private_data_len = sizeof(request_data);
    param.retry_count = 7;
    param.rnr_retry_count = 7;

    request_data.qid = htole16(queue->qid);
    request_data.hrqsize = htole16(queue->queue_depth);
    request_data.hsqsize = htole16(queue->queue_depth - 1);

    nvmx_debug("Connect using RDMA device %s",
               queue->cm_id->verbs->device->name);
    ret = rdma_connect(queue->cm_id, &param);
    if (ret) {
        nvmx_error("nvme rdma connect error");
	goto out;
    }

    return 0;

out:
    return ret;
}

/*
 * Must be called with state_lock held
 */
static int nvmf_rdma_connect(nvmf_rdma_qpair_t *queue)
{
    struct rdma_cm_event *event;
    struct nvmf_rdma_accept_private_data *accept_data;
    int err = 0;

    if (_nvmf_rdma_connect(queue)) {
        nvmx_error("queue %d: failed connecting", queue->qid);
        return -1;
    }

    event = nvmf_rdma_get_cm_event(queue->cm_channel, RDMA_CM_EVENT_ESTABLISHED,
                                   NVME_RDMA_TIMEOUT_IN_MS);

    if (event == NULL) {
        nvmx_error("queue %d: failed to receive expected event", queue->qid);
        return -1;
    }

    accept_data = (struct nvmf_rdma_accept_private_data *)
                  event->param.conn.private_data;
    if (!accept_data) {
        nvmx_error("queue %d: NVMe-oF target did not return accept data",
                   queue->qid);
        err = -1;
        goto out;
    }

    nvmx_debug("queue %d: requested queue depth %d. Actually got queue depth %d.",
               queue->qid, queue->queue_depth, accept_data->crqsize);

    queue->queue_depth = nvmx_min(queue->queue_depth, accept_data->crqsize);

out:
    rdma_ack_cm_event(event);
    return err;
}

/*
 * Must be called with state_lock held
 */
static int nvmf_rdma_alloc_reqs_rsps(nvmf_rdma_qpair_t *queue)
{
    int rc;

    rc = nvmf_rdma_alloc_reqs(queue);
    if (rc) {
        nvmx_error("queue %d: unable to allocate RDMA requests for queue",
                   queue->qid);
	goto out;
    }

    rc = nvmf_rdma_alloc_rsps(queue);
    if (rc) {
        nvmx_error("queue %d: unable to allocate RDMA responses for queue",
                   queue->qid);
        goto free_reqs;
    }

    queue->state = NVMF_RDMA_QUEUE_READY;

    return 0;

free_reqs:
    nvmf_rdma_free_reqs(queue);
out:
    return rc;
}

static int nvmf_rdma_parse_addr(struct sockaddr_storage *sa,
                                enum nvmf_adrfam adrfam,
                                const char *addr, const char *svcid)
{
    struct addrinfo *res;
    struct addrinfo hints;
    int ret;

    memset(&hints, 0, sizeof(hints));
    if (adrfam == NVMF_ADRFAM_IPV4)
        hints.ai_family = AF_INET;
    else if (adrfam == NVMF_ADRFAM_IPV6)
        hints.ai_family = AF_INET6;
    else
        return EINVAL;

    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = 0;

    ret = getaddrinfo(addr, svcid, &hints, &res);
    if (ret) {
	nvmx_error("getaddrinfo failed: %s (%d)", gai_strerror(ret), ret);
        goto out;
    }

    if (res->ai_addrlen > sizeof(*sa)) {
        nvmx_error("ai_addrlen %zu too large", (size_t)res->ai_addrlen);
        ret = EINVAL;
    } else {
        memcpy(sa, res->ai_addr, res->ai_addrlen);
    }

    freeaddrinfo(res);
out:
    return ret;

}

static void nvmf_rdma_deactivate_queue(nvmf_emu_ctrl_t *fctrl, uint16_t qid)
{
    nvmf_rdma_ctrl_t *ctrl = to_rdma_ctrl(fctrl);
    nvmf_rdma_qpair_t *queue = &ctrl->queues[qid];
    int rc;

    pthread_mutex_lock(&queue->state_lock);
    if (queue->state != NVMF_RDMA_QUEUE_CONNECTED) {
        nvmx_debug("queue %d: Can't deactivate. Might be already deactivated",
                   qid);
        goto out;
    }

    nvmx_debug("queue %d: deactivating for ctrl 0x%p", qid, ctrl);

    if (queue->offload_type == NVMD_RDMA_OFFLOAD_SNAPv3) {
        nvme_emulator_t *dev = queue->ctrl->ctrl.ep->io_driver->dev;

        rc = nvme_emu_detach_sq_qp(dev, qid);
        if (rc)
            goto out;
    }

    queue->state = NVMF_RDMA_QUEUE_READY;
out:
    pthread_mutex_unlock(&queue->state_lock);
}

/* Called with state_lock taken */
static int nvmf_rdma_qpair_connect(nvmf_rdma_qpair_t *queue)
{
    int rc;

    rc = nvme_fabric_qpair_connect(&queue->ctrl->ctrl, queue->qid,
                                   queue->queue_depth);
    if (rc) {
        nvmx_debug("queue %d: qpair connect failed", queue->qid);
        return rc;
    }

    if (queue->offload_type == NVMD_RDMA_OFFLOAD_SNAPv3) {
        nvme_emulator_t *dev = queue->ctrl->ctrl.ep->io_driver->dev;

        rc = nvme_emu_attach_sq_qp(dev, queue->qid, queue->qp);
        if (rc)
            return rc;
    }
    queue->state = NVMF_RDMA_QUEUE_CONNECTED;
    nvmx_debug("queue %d: connected successfully", queue->qid);

    return rc;
}

static int nvmf_rdma_activate_queue(nvmf_emu_ctrl_t *fctrl, uint16_t qid)
{
    int rc = 0;
    nvmf_rdma_ctrl_t *ctrl = to_rdma_ctrl(fctrl);
    nvmf_rdma_qpair_t *queue = &ctrl->queues[qid];

    pthread_mutex_lock(&queue->state_lock);
    switch (queue->state) {
    case NVMF_RDMA_QUEUE_CONNECTED:
        nvmx_debug("queue %d: is already activated", qid);
        rc = -EALREADY;
        break;
    case NVMF_RDMA_QUEUE_DISCONNECTED:
        nvmx_error("queue %d: is not ready", qid);
        rc = -ENOTCONN;
        break;
    case NVMF_RDMA_QUEUE_READY:
        nvmx_debug("queue %d: activating for ctrl 0x%p", qid, ctrl);
        rc = nvmf_rdma_qpair_connect(queue);
        break;
    default:
        nvmx_error("queue %d: is in unknown %d state", qid, queue->state);
        rc = -EINVAL;
        break;
    }
    pthread_mutex_unlock(&queue->state_lock);

    return rc;
}

static int nvmf_rdma_destroy_queue(nvmf_emu_ctrl_t *fctrl, uint16_t qid)
{
    nvmf_rdma_ctrl_t *ctrl = to_rdma_ctrl(fctrl);
    nvmf_rdma_qpair_t *queue = &ctrl->queues[qid];

    pthread_mutex_lock(&queue->state_lock);
    if (queue->state == NVMF_RDMA_QUEUE_DISCONNECTED) {
        nvmx_debug("queue %d: Can't destroy. Might be already destroyed", qid);
        goto out_unlock;
    }

    nvmx_debug("queue %d: destroying for ctrl 0x%p", qid, ctrl);

    rdma_disconnect(queue->cm_id);
    nvmf_rdma_destroy_transport_resources(queue);
    rdma_destroy_id(queue->cm_id);
    /*
     * RDMA requests/responses are allocated only after ESTABLISHED rdma_cm
     * event is detected, after final queue_depth value is determined.
     * In contrary, we shouldn't free them until all in-flight requests
     * were handled, and this can be assured only after rdma_destroy_id().
     */
    nvmf_rdma_free_rsps(queue);
    nvmf_rdma_free_reqs(queue);

    rdma_destroy_event_channel(queue->cm_channel);

    queue->state = NVMF_RDMA_QUEUE_DISCONNECTED;
out_unlock:
    pthread_mutex_unlock(&queue->state_lock);

    return 0;
}

/*
 * Must be called with state_lock held
 */
static int nvmf_rdma_resolve_addr(nvmf_rdma_qpair_t* queue,
                                  struct sockaddr_storage *src_addr,
                                  struct sockaddr_storage *dst_addr)
{
    int rc;
    struct rdma_cm_event *event;

    rc = rdma_resolve_addr(queue->cm_id, (struct sockaddr *)src_addr,
                           (struct sockaddr *)dst_addr,
                           NVME_RDMA_TIMEOUT_IN_MS);
    if (rc) {
        nvmx_error("queue %d: rdma_resolve_addr failed with error: %d",
                   queue->qid, errno);
        return rc;
    }

    event = nvmf_rdma_get_cm_event(queue->cm_channel,
                                   RDMA_CM_EVENT_ADDR_RESOLVED, -1);
    if (event == NULL) {
        nvmx_error("queue %d: failed to receive expected event", queue->qid);
        return -1;
    }
    rdma_ack_cm_event(event);

    return 0;
}


static int nvmf_rdma_create_queue(nvmf_emu_ctrl_t *fctrl, uint16_t qid,
                                  uint16_t queue_depth)
{
    nvmf_rdma_ctrl_t *ctrl = to_rdma_ctrl(fctrl);
    nvmf_rdma_qpair_t *queue = &ctrl->queues[qid];
    bool src_addr_specified = false;
    struct sockaddr_storage dst_addr;
    struct sockaddr_storage src_addr;
    int rc = -1;

    pthread_mutex_lock(&queue->state_lock);
    if (queue->state != NVMF_RDMA_QUEUE_DISCONNECTED) {
        nvmx_debug("queue %d: is already created", qid);
        rc = -EALREADY;
        goto out;
    }

    nvmx_debug("queue %d: creating for ctrl 0x%p", qid, ctrl);

    queue->queue_depth = queue_depth;
    queue->cm_channel = rdma_create_event_channel();
    if (!queue->cm_channel) {
        nvmx_error("queue %d: rdma_create_event_channel() failed", qid);
        goto out;
    }
    if (nvmf_rdma_set_channel_fd_nonblock(queue->cm_channel)) {
        nvmx_error("queue %d: failed to set channel fd to non blocking", qid);
        goto destroy_channel;
    }

    rc = rdma_create_id(queue->cm_channel, &queue->cm_id, queue, RDMA_PS_TCP);
    if (rc) {
        nvmx_error("queue %d: rdma_create_id() failed", qid);
        goto destroy_channel;
    }

    memset(&dst_addr, 0, sizeof(dst_addr));
    rc = nvmf_rdma_parse_addr(&dst_addr, ctrl->ctrl.id.adrfam,
                              ctrl->ctrl.id.traddr, ctrl->ctrl.id.trsvcid);
    if (rc) {
        nvmx_error("queue %d: dst_addr nvmf_rdma_parse_addr() failed", qid);
        goto destroy_id;
    }

    if (ctrl->ctrl.src_addr[0]) {
        memset(&src_addr, 0, sizeof(src_addr));
        rc = nvmf_rdma_parse_addr(&src_addr, ctrl->ctrl.id.adrfam,
                                  ctrl->ctrl.src_addr, NULL);
        if (rc) {
            nvmx_error("queue %d: src_addr nvmf_rdma_parse_addr() failed",
                       qid);
            goto destroy_id;
        }
        src_addr_specified = true;
    }

    rc = nvmf_rdma_resolve_addr(queue, src_addr_specified ? &src_addr : NULL,
                                &dst_addr);
    if (rc)
        goto destroy_id;

    rc = nvmf_rdma_resolve_route(queue);
    if (rc)
        goto destroy_id;

    rc = nvmf_rdma_connect(queue);
    if (rc)
        goto destroy_transport;

    rc = nvmf_rdma_alloc_reqs_rsps(queue);
    if (rc) {
        nvmx_error("queue %d: failed allocating reqs and rsps", qid);
        goto disconnect;
    }

    pthread_mutex_unlock(&queue->state_lock);
    nvmx_debug("queue %d: RDMA connection was established successfully", qid);

    return 0;

disconnect:
    rdma_disconnect(queue->cm_id);
destroy_transport:
    nvmf_rdma_destroy_transport_resources(queue);
destroy_id:
    rdma_destroy_id(queue->cm_id);
destroy_channel:
    rdma_destroy_event_channel(queue->cm_channel);
out:
    pthread_mutex_unlock(&queue->state_lock);
    return rc;
}

/*
 * Build SGL describing empty payload.
 */
static int nvmf_rdma_build_null_request(struct nvmf_rdma_req *req)
{
    nvme_sgl_desc_t *ksgl;

    req->nvmf_req.cmd.common.psdt = NVME_PSDT_SGL_MPTR_CONTIG;

    ksgl = &req->nvmf_req.cmd.common.dptr.sgl;
    ksgl->type = NVME_SGL_TYPE_KEYED_DATA_BLOCK;
    ksgl->subtype = NVME_SGL_SUBTYPE_ADDRESS;
    ksgl->keyed_data_block.length = 0;
    ksgl->keyed_data_block.key = 0;
    ksgl->keyed_data_block.address = 0;

    return 0;
}

/*
 * Build SGL describing contiguous payload buffer.
 */
static int nvmf_rdma_build_request(struct nvmf_rdma_req *req)
{
    void *payload = req->nvmf_req.payload;
    struct ibv_mr *mr = req->payload_mr;
    nvme_sgl_desc_t *ksgl;

    if (req->nvmf_req.payload_size == 0)
        return EINVAL;

    req->nvmf_req.cmd.common.psdt = NVME_PSDT_SGL_MPTR_CONTIG;

    ksgl = &req->nvmf_req.cmd.common.dptr.sgl;

    ksgl->type = NVME_SGL_TYPE_KEYED_DATA_BLOCK;
    ksgl->subtype = NVME_SGL_SUBTYPE_ADDRESS;
    ksgl->keyed_data_block.length = req->nvmf_req.payload_size;
    ksgl->keyed_data_block.key = mr->rkey;
    ksgl->keyed_data_block.address = htole64((uint64_t)payload);

    return 0;
}

static int nvmf_rdma_req_init(nvmf_rdma_qpair_t *queue, void *data,
                              int data_len, struct nvmf_rdma_req *req)
{
    int rc;

    req->nvmf_req.cmd.common.cid = req->nvmf_req.id;
    if (data_len <= NVMF_RDMA_PAYLOAD_BUFFER_SIZE)
        req->nvmf_req.payload_size = data_len;
    else
        return -1;


    if (req->nvmf_req.payload != data && data_len)
        memcpy(req->nvmf_req.payload, data, data_len);


    if (req->nvmf_req.payload_size == 0)
        rc = nvmf_rdma_build_null_request(req);
    else
        rc = nvmf_rdma_build_request(req);

    return rc;
}

/*
 * cmd should be embedded in rdma_req
 */
static int nvmf_rdma_send_cmd_async(nvmf_emu_ctrl_t *ctrl, uint16_t qid,
                                    nvmf_req_t *nvmf_req, void *data, int data_len)
{
    nvmf_rdma_ctrl_t *rctrl = to_rdma_ctrl(ctrl);
    nvmf_rdma_qpair_t *queue = &rctrl->queues[qid];
    struct nvmf_rdma_req *rdma_req = to_rdma_req(nvmf_req);
    struct ibv_send_wr *bad_wr = NULL;
    int ret;

    /* TODO: check if queue if ready to send CC */
    ret = nvmf_rdma_req_init(queue, data, data_len, rdma_req);
    if (unlikely(ret)) {
        nvmx_error("queue %d: failed to init request", qid);
        return ret;
    }

    /*TODO: add timeout mechanism*/
    rdma_req->nvmf_req.timed_out = false;

    ret = ibv_post_send(queue->qp, &rdma_req->send_wr, &bad_wr);
    if (unlikely(ret)) {
        nvmx_error("queue %d: failed to post a send request", qid);
        return ret;
    }

    nvmx_debug("queue %d: post a send request with id %d opcode %u",
               qid, nvmf_req->id, nvmf_req->cmd.common.opc);

    return 0;
}

static int nvmf_rdma_send_cmd_sync(nvmf_emu_ctrl_t *ctrl, uint16_t qid,
                                   nvmf_req_t *nvmf_req, void *data, int data_len,
                                   nvmf_completion_status_t *status)
{
    struct nvmf_rdma_req *rdma_req = to_rdma_req(nvmf_req);
    int ret;

    rdma_req->nvmf_req.status = status;

    ret = nvmf_rdma_send_cmd_async(ctrl, qid, nvmf_req, data, data_len);
    if (unlikely(ret))
        return ret;

    sem_wait(&rdma_req->nvmf_req.sync);

    return 0;
}

static void nvmf_rdma_uninit_queue(nvmf_rdma_qpair_t *queue)
{
    pthread_mutex_destroy(&queue->state_lock);
}

static int nvmf_rdma_init_queue(nvmf_rdma_ctrl_t *ctrl, int qid)
{
    nvmf_rdma_qpair_t *queue = &ctrl->queues[qid];
    nvme_emu_driver_t *io_driver = ctrl->ctrl.ep->io_driver;
    enum nvmf_rdma_offload offload_type = NVMF_RDMA_OFFLOAD_NONE;
    int err;

    if (qid && io_driver && io_driver->dev) {
        if (io_driver->dev->io_q_ops->q_type == NVME_EMU_QUEUE_NVME2NVMF_REMOTE) {
            if (strcmp(io_driver->dev->name, SNAP_EMU_NAME) == 0) {
                offload_type = NVMD_RDMA_OFFLOAD_SNAPv3;
            } else {
                nvmx_error("queue %d: %s emulator does not support offload mode",
                           io_driver->dev->name);
                return NVMX_ERR;
            }
        }
    }

    queue->qid = qid;
    queue->ctrl = ctrl;
    queue->cmnd_capsule_len = sizeof(nvme_cmd_t);
    queue->offload_type = offload_type;

    err = pthread_mutex_init(&queue->state_lock, NULL);
    return err;
}

static void nvmf_rdma_free_queues(nvmf_rdma_ctrl_t *ctrl)
{
    int qid;

    for (qid = 0; qid < ctrl->ctrl.nr_queues; qid++)
        nvmf_rdma_uninit_queue(&ctrl->queues[qid]);

    free(ctrl->queues);
}

void nvmf_rdma_free_ctrl(nvmf_emu_ctrl_t *ctrl)
{
    nvmf_rdma_ctrl_t *rctrl = to_rdma_ctrl(ctrl);

    nvmf_rdma_free_queues(rctrl);
    free(rctrl);
}

static int nvmf_rdma_alloc_queues(nvmf_rdma_ctrl_t *ctrl)
{
    int qid, ret = -ENOMEM;
    int i;

    ctrl->queues = calloc(ctrl->ctrl.nr_queues, sizeof(*ctrl->queues));
    if (!ctrl->queues)
        goto err;

    for (qid = 0; qid < ctrl->ctrl.nr_queues; qid++) {
        ret = nvmf_rdma_init_queue(ctrl, qid);
        if (ret)
            goto err_uninit_queues;
    }

    return 0;

err_uninit_queues:
    for (i = 0; i < qid; i++)
        nvmf_rdma_uninit_queue(&ctrl->queues[i]);
    free(ctrl->queues);
err:
    return ret;
}

nvmf_emu_ctrl_t *nvmf_rdma_alloc_ctrl(nvme_emu_ep_t *ep)
{
    nvmf_rdma_ctrl_t *ctrl;
    int err;

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        goto out;

    ctrl->ctrl.ep = ep;
    ctrl->ctrl.nr_queues = ep->io_driver->num_queues;
    ctrl->ctrl.cmd_ops.send_cmd_sync = nvmf_rdma_send_cmd_sync;
    ctrl->ctrl.cmd_ops.send_cmd_async = nvmf_rdma_send_cmd_async;
    ctrl->ctrl.cmd_ops.get_request = nvmf_rdma_get_request;
    ctrl->ctrl.cmd_ops.put_request = nvmf_rdma_put_request;
    ctrl->ctrl.q_ops.create = nvmf_rdma_create_queue;
    ctrl->ctrl.q_ops.destroy = nvmf_rdma_destroy_queue;
    ctrl->ctrl.q_ops.activate = nvmf_rdma_activate_queue;
    ctrl->ctrl.q_ops.deactivate = nvmf_rdma_deactivate_queue;
    ctrl->ctrl.max_io_queue_size = NVMF_MAX_IO_QUEUE_SIZE;

    err = nvmf_rdma_alloc_queues(ctrl);
    if (err)
        goto free_ctrl;

    nvmx_info("NVMf/RDMA transport ctrl allocated");

    return &ctrl->ctrl;

free_ctrl:
    free(ctrl);
out:
    return NULL;
}
