#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "compiler.h"
#include "mlnx_atomic.h"
#include "nvme_emu_log.h"
#include "lock_free_stack.h"
#include "resman.h"

int resman_init(resman_t *resman, size_t thread_count)
{
    memset(resman, 0, sizeof(resman_t));

    if (!thread_count) {
        return -EINVAL;
    }

    resman->thread_count = thread_count;
    resman->threads = calloc(thread_count, sizeof(resman_thread_ctx_t));
    if (!resman->threads)
        goto err;

    resman->levels = calloc(thread_count, sizeof(resman->levels[0]));
    if (!resman->levels)
        goto free_threads;

    return 0;

free_threads:
    free(resman->threads);
err:
    return -ENOMEM;
}

void resman_clear(resman_t *resman)
{
    free(resman->levels);
    free(resman->threads);
}

int resman_add(resman_t *resman, resman_entry_t *entry)
{
    // Add new entry to global stack
    lfs_push_entry(&resman->stack, &entry->lfs_entry);
    resman->len++;

    return 0;
}

static inline size_t resman_get_quota(resman_t *resman, size_t thread_id)
{
    size_t quota, rem;
    const size_t total_demand = resman->total_demand;
    resman_thread_ctx_t *ctx = &resman->threads[thread_id];

    if (!total_demand)
        return 0;

    // The actual formula for quota is
    // quota = ROUND_UP ((ctx->demand / total_demand) * resman->len)
    // But since ctx->demand is already pre-multiplied by resman->len
    // We use simplified math here
    quota = ctx->demand / total_demand;
    rem = ctx->demand % total_demand;

    return quota + (rem ? 1 : 0);
}

static inline int resman_get_balance(resman_t *resman, size_t thread_id)
{
    // Balance is negative if thread got less than desired
    // If it wants 5 entries, but got only 3, the balance is -2
    // The fully balanced state is when each thread has non-negative balance
    const size_t quota = resman_get_quota(resman, thread_id);
    return resman->levels[thread_id] - quota;
}

static inline bool resman_claim_entries(resman_t *resman)
{
    size_t thread_no;
    lfs_entry_t *entry;
    bool ret = false;

    // Go through all the threads and return everything above quota to global
    // pool. This might be done just partially because some or even all entries
    // above quota may be in use. Final balancing will be done in resman_put
    for (thread_no = 0; thread_no < resman->thread_count; thread_no++) {
        while (resman_get_balance(resman, thread_no) > 0) {
            entry = lfs_pop_entry(&resman->threads[thread_no].stack);
            if (entry) {
                atomic_fetch_sub(&resman->levels[thread_no], 1);
                lfs_push_entry(&resman->stack, entry);
                // Return true if something was returned to the global pool
                ret = true;
            }
            else
                break;
        }
    }

    return ret;
}

void resman_inc_demand(resman_t *resman, size_t thread_id)
{
    resman_thread_ctx_t *ctx = &resman->threads[thread_id];

    // This API allows only to change demand sequentially
    // Hence we can pretend that we pre-multiply demand by resman->len
    // in order to simplify further math
    ctx->demand += resman->len;

    // Adjust total_demand
    atomic_fetch_add(&resman->total_demand, 1);
}

void resman_dec_demand(resman_t *resman, size_t thread_id)
{
    resman_thread_ctx_t *ctx = &resman->threads[thread_id];

    ctx->demand -= resman->len;

    // Adjust total_demand
    atomic_fetch_sub(&resman->total_demand, 1);
}

resman_entry_t *resman_get(resman_t *resman, size_t thread_id)
{
    lfs_entry_t *entry;

    // If we have local entry - use it
    entry = lfs_pop_entry(&resman->threads[thread_id].stack);

    // If all local entries are in use and we didn't reach our quota
    // try to get some from global pool
    if (!entry && resman_get_balance(resman, thread_id) < 0) {

        // pop from global pool
        entry = lfs_pop_entry(&resman->stack);

        // If global pool is also empty - claim entries from other threads
        if (!entry) {
            // claimed items (if any) will appear in global pool
            if (resman_claim_entries(resman)) {
                // Try global pool another time
                entry = lfs_pop_entry(&resman->stack);
            }
        }

        if (entry)
            atomic_fetch_add(&resman->levels[thread_id], 1);
    }

    return entry ? container_of(entry, resman_entry_t, lfs_entry) : NULL;
}

bool resman_put(resman_t *resman, size_t thread_id, resman_entry_t *entry)
{
    // If we didn't exceed our quota - return entry to local pool
    if (resman_get_balance(resman, thread_id) <= 0) {
        lfs_push_entry(&resman->threads[thread_id].stack, &entry->lfs_entry);
        return true;
    }
    else {
        // Otherwise, push entry to global pool
        lfs_push_entry(&resman->stack, &entry->lfs_entry);
        atomic_fetch_sub(&resman->levels[thread_id], 1);
        return false;
    }
}

bool resman_quota_exceed(resman_t *resman, size_t thread_id)
{
    return resman_get_balance(resman, thread_id) > 0;
}

