/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdlib.h>
#include <stdbool.h>
#include <snap.h>
#include "spdk_ext_io.h"
#ifdef HAVE_RTE_HASH_H
#include <rte_hash.h>
#endif

int spdk_ext_io_zcopy_memory_domain_translate_memory_cb(
        struct spdk_memory_domain *src_domain,
        spdk_ext_io_zcopy_ctx_t *zcopy, struct spdk_memory_domain *dst_domain,
        struct spdk_memory_domain_translation_ctx *dst_domain_ctx,
        void *addr, size_t len,
        struct spdk_memory_domain_translation_result *result)
{
    struct ibv_qp *src_qp;
    struct snap_cross_mkey *mkey;

    if (unlikely(!zcopy->enabled))
       return -1;

    if (unlikely(!dst_domain_ctx || !dst_domain_ctx->rdma.ibv_qp))
       return -1;

    src_qp = (struct ibv_qp *)dst_domain_ctx->rdma.ibv_qp;

    if (likely(zcopy->pd == src_qp->pd)) {
        mkey = zcopy->mkey;
    } else {
        mkey = zcopy->find_key(zcopy, src_qp->pd);
        if (unlikely(!mkey))
            return -1;

        zcopy->pd = src_qp->pd;
        zcopy->mkey = mkey;
    }

#ifdef HAVE_STRUCT_SPDK_MEMORY_DOMAIN_TRANSLATION_RESULT_IOV
    result->iov.iov_len = len;
    result->iov.iov_base = addr;
    result->iov_count = 1;
#else
    result->len = len;
    result->addr = addr;
#endif
    result->rdma.lkey = mkey->mkey;
    result->rdma.rkey = mkey->mkey;
    result->dst_domain = dst_domain;

    return 0;
}

int spdk_ext_io_zcopy_status(struct spdk_bdev *bdev)
{
    return spdk_ext_io_get_bdev_status(bdev, SPDK_DMA_DEVICE_TYPE_RDMA);
}

#ifdef HAVE_RTE_HASH_H

/*
 * find cross mkey from key_table by use @pd, if it does not exist,
 *  then create a cross mkey based on this @pd and add to key_table.
 **/
static struct snap_cross_mkey *spdk_ext_io_zcopy_find_key(
                spdk_ext_io_zcopy_ctx_t *zcopy_ctx,
                struct ibv_pd *pd)
{
	int ret;
    struct snap_cross_mkey *mkey = NULL;
    struct snap_cross_mkey_attr cm_attr = {};

    ret = rte_hash_lookup_data(zcopy_ctx->key_table, (const void *)pd, (void **)&mkey);
    if (likely(ret >= 0))
        return mkey;

    if (ret == -EINVAL)
        return NULL;

    cm_attr.vtunnel = zcopy_ctx->sdev->mdev.vtunnel;
    cm_attr.dma_rkey = zcopy_ctx->sdev->dma_rkey;
    cm_attr.vhca_id = snap_get_vhca_id(zcopy_ctx->sdev);
    cm_attr.crossed_vhca_mkey = zcopy_ctx->sdev->crossed_vhca_mkey;

    mkey = snap_create_cross_mkey_by_attr(pd, &cm_attr);
    if (!mkey)
        return NULL;

    ret = rte_hash_add_key_data(zcopy_ctx->key_table, pd, mkey);
    if (ret)
        goto destroy_mkey;

    return mkey;

destroy_mkey:
    snap_destroy_cross_mkey(mkey);

    return NULL;
}

void spdk_ext_io_zcopy_init(spdk_ext_io_zcopy_ctx_t *zcopy_ctx,
                            void *key_table,
                            struct snap_device *sdev)
{
    zcopy_ctx->enabled = true;
    zcopy_ctx->sdev = sdev;
    zcopy_ctx->find_key = spdk_ext_io_zcopy_find_key;
    zcopy_ctx->key_table = key_table;
    zcopy_ctx->pd = NULL;
    zcopy_ctx->mkey = NULL;
}

void spdk_ext_io_zcopy_clear(spdk_ext_io_zcopy_ctx_t *zcopy_ctx)
{
    uint32_t next = 0;
    struct ibv_pd *pd;
    struct snap_cross_mkey *mkey;

    if (!zcopy_ctx->key_table)
        return;

    if (rte_hash_count(zcopy_ctx->key_table) == 0)
        return;

    /* try the best to delete pd-mkey pair from hash table */
    while (rte_hash_iterate(zcopy_ctx->key_table, (const void **)&pd, (void **)&mkey, &next) >= 0) {
        if (rte_hash_del_key(zcopy_ctx->key_table, (const void *)pd) < 0)
            continue;

        snap_destroy_cross_mkey(mkey);
    };
}

#else

void spdk_ext_io_zcopy_init(spdk_ext_io_zcopy_ctx_t *zcopy_ctx,
                            void *key_table,
                            struct snap_device *sdev)
{
    return;
}

void spdk_ext_io_zcopy_clear(spdk_ext_io_zcopy_ctx_t *zcopy_ctx)
{
    return;
}

#endif
