/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "zcopy_snap.h"
#include "nvme_emu_log.h"

#ifdef HAVE_RTE_HASH_H

#include <infiniband/verbs.h>
#include <rte_hash.h>
#include <rte_jhash.h>
#include <rte_errno.h>

#define ZCOPY_MIN_HASH_TABLE_SIZE    8

void zcopy_snap_destroy_key_table(void *key_table)
{
    rte_hash_free((struct rte_hash *)key_table);
}

void *zcopy_snap_create_key_table(char *name)
{
    int dev_num;
    struct ibv_device **list;
    struct rte_hash *key_table;
    struct rte_hash_parameters params = {0};

    list = ibv_get_device_list(&dev_num);
    if (!list) {
        nvmx_error("failed to open IB device list");
        return NULL;
    }

    ibv_free_device_list(list);

    if (unlikely(dev_num == 0)) {
        nvmx_info("No IB device exist");
        return NULL;
    }

    params.entries = dev_num >= ZCOPY_MIN_HASH_TABLE_SIZE ? dev_num : ZCOPY_MIN_HASH_TABLE_SIZE;
    params.key_len = sizeof(struct ibv_pd *);
    params.hash_func = rte_jhash;
    params.name = name;
    params.extra_flag |= RTE_HASH_EXTRA_FLAGS_RW_CONCURRENCY_LF;

    key_table = rte_hash_create(&params);
    if (key_table)
        rte_hash_reset(key_table);
    else
        nvmx_error("rte_hash_create() failed for table:%s, rte_errno:%d", name, rte_errno);

    return (void *)key_table;
}

#else

void zcopy_snap_destroy_key_table(void *key_table)
{
    return;
}
void *zcopy_snap_create_key_table(char *name)
{
    return NULL;
}

#endif
